import SDKAdapter from '../core/adapter.js';
import { SDKCollection } from '../core/collection.js';
import { SDKModel } from '../core/model.js';
import { SDKModelSchema } from '../core/schema.js';
import type { SDK } from '../sdk.js';
import { Atleast, RequiredKeys } from '../utils/typescript.js';
import type { ComponentModel } from './components.js';
import { LibraryModel, NanoId } from './libraries.js';
/** Defines the schema for Collection. */
export interface Collection {
    /** The identifier of the Library that Collection belongs to */
    libraryId: NanoId;
    /** The identifier of the Collection */
    id: NanoId;
    /**
     * The name of the Collection.
     *
     * @minLength 1
     * @maxLength 128
     */
    name: string;
    /** Specifies whether the Collection is the default one or not, each Library can have only one default Collection */
    isDefault: boolean;
    /**
     * Specifies the order of appearance of the Collection.
     *
     * @minLength 0
     * @maxLength 64
     * @nullable
     */
    orderIdx?: string;
    /** The timestamp the Collection was created at */
    createdAt?: Date;
    /** The timestamp when the Collection was last modified at */
    modifiedAt?: Date;
}
export declare class Collections extends SDKAdapter<Collection> {
    fetch(collection: Pick<CollectionParams, 'libraryId'>): Promise<Collection[]>;
    get(collection: Pick<CollectionParams, 'libraryId' | 'id'>): Promise<Collection>;
    post(collection: Atleast<CollectionParams, 'libraryId'>): Promise<Collection>;
    put(collection: Atleast<CollectionParams, 'libraryId' | 'id'>): Promise<Collection>;
    delete(collection: Pick<CollectionParams, 'libraryId' | 'id'>): Promise<Collection>;
}
export declare function getCollectionDefaults(sdk: SDK): {
    id: string;
    isDefault: boolean;
    orderIdx: string;
    createdAt: Date;
    modifiedAt: Date;
};
export type CollectionMinimal<T extends keyof Collection = never> = RequiredKeys<Collection, typeof getCollectionDefaults, T>;
export interface CollectionImplicit {
    libraryId: LibraryModel['id'];
    library?: LibraryModel;
    components?: Partial<ComponentModel>[];
}
export interface CollectionParams extends Collection, CollectionImplicit {
}
export interface CollectionModel extends CollectionParams {
}
/**
 * Represents a CollectionModel, a groupping of components.
 * This class provides methods and properties for working with collections.
 */
export declare class CollectionModel extends SDKModel<Collection, CollectionMinimal, CollectionImplicit> implements CollectionParams {
    /**
     * Gets the adapter for the collection.
     */
    get adapter(): Collections;
    components: ReturnType<this['constructComponents']>;
    /**
     * Defines additional properties for the collection.
     */
    defineProperties(): void;
    /**
     * Constructs the components for the collection.
     *
     * @returns The constructed components SDK collection.
     */
    constructComponents(): SDKCollection<typeof ComponentModel, ComponentModel, import("./components.js").Component, Omit<Pick<Partial<import("./components.js").Component>, "description" | "id" | "status" | "createdAt" | "modifiedAt" | "versionCount" | "stagedVersionCount" | "embeddableVersionCount" | "datasourceIds" | "modifiedBy"> & Omit<import("./components.js").Component, "description" | "id" | "status" | "createdAt" | "modifiedAt" | "versionCount" | "stagedVersionCount" | "embeddableVersionCount" | "datasourceIds" | "modifiedBy"> & Partial<import("./components.js").ComponentImplicit>, "collection" | "collectionId" | "libraryId" | "library">, {
        id: string;
    } | {
        details: {
            id: string;
        };
    } | {
        id: string;
        revision: number;
        status: "published" | "draft" | "staged" | "saved";
    }>;
    /**
     * Gets the hidden properties of the collection. These properties will be made non-iterable.
     *
     * @returns An array of hidden properties.
     */
    getHiddenProps(): string[];
    /**
     * Gets the default values for the collection. These values will be used if not provided.
     *
     * @param sdk  - The SDK instance.
     * @returns The default values.
     */
    getDefaults(sdk: SDK): {
        id: string;
        isDefault: boolean;
        orderIdx: string;
        createdAt: Date;
        modifiedAt: Date;
    };
    /**
     * Compares the collection with another collection for sorting purposes.
     *
     * @param other  - The other collection to compare.
     * @returns A number indicating the sort order.
     */
    sortCompare(other: this): number;
    /**
     * Checks if the collection matches the given search string.
     *
     * @param search  - The search string.
     * @returns True if the collection matches the search string, false otherwise.
     */
    matchesSearch(search: string): boolean;
    /**
     * Gets the path of the collection.
     *
     * @returns The path of the collection.
     */
    getPath(): string;
    serialize(): {
        components: {
            versions: import("./versions.js").Version[];
            bundles: import("./version-bundles.js").VersionBundle[];
            libraryId: string;
            id: string;
            collectionId: string;
            name: string;
            description?: string;
            status: "published" | "deleted" | "draft" | "staged" | "changed";
            versionCount?: number;
            stagedVersionCount?: number;
            embeddableVersionCount?: number;
            datasourceIds?: string[];
            createdAt?: Date;
            stagedAt?: Date;
            publishedAt?: Date;
            modifiedAt?: Date;
            modifiedBy?: object;
        }[];
        /** The identifier of the Library that Collection belongs to */
        libraryId: string;
        /** The identifier of the Collection */
        id: string;
        /**
         * The name of the Collection.
         *
         * @minLength 1
         * @maxLength 128
         */
        name: string;
        /** Specifies whether the Collection is the default one or not, each Library can have only one default Collection */
        isDefault: boolean;
        /**
         * Specifies the order of appearance of the Collection.
         *
         * @minLength 0
         * @maxLength 64
         * @nullable
         */
        orderIdx?: string;
        /** The timestamp the Collection was created at */
        createdAt?: Date;
        /** The timestamp when the Collection was last modified at */
        modifiedAt?: Date;
    };
    /**
     * Copies the collection (with components by default) to another library.
     *
     * @param target             - The target library to copy the collection to.
     * @param includeComponents  - Optional. Specifies whether to include the components in the copy. Default is true.
     * @returns A promise that resolves to the copied collection.
     */
    copyToLibrary(target: LibraryModel, includeComponents?: boolean): Promise<CollectionModel>;
    /**
     * Gets the schema for the CollectionModel class.
     *
     * @returns The schema for the CollectionModel class.
     */
    static get schema(): SDKModelSchema<typeof CollectionModel, Collection>;
}
//# sourceMappingURL=collections.d.ts.map