import SDKAdapter from '../core/adapter.js';
import { SDKCollection } from '../core/collection.js';
import { SDKModel } from '../core/model.js';
import { SDKModelSchema } from '../core/schema.js';
import type { SDK } from '../sdk.js';
import { Atleast, RequiredKeys } from '../utils/typescript.js';
import type { CollectionModel } from './collections.js';
import { DataOptionsById, DatasourceModel } from './datasources.js';
import type { LibraryModel, NanoId } from './libraries.js';
import { Version, VersionModel } from './versions.js';
/**
 * Defines the schema for Component.
 *
 * @category Component2
 * @module Component1
 */
export interface Component {
    /** The identifier of the Library that Component belongs to */
    libraryId: NanoId;
    /** The identifier of the Component */
    id: NanoId;
    /** The identifier of the Collection */
    collectionId: NanoId;
    /**
     * The name of the Component.
     *
     * @minLength 1
     * @maxLength 128
     */
    name: string;
    /**
     * The description of the Component.
     *
     * @minLength 0
     * @maxLength 128
     */
    description?: string;
    /** The status of the Component */
    status: 'published' | 'deleted' | 'staged' | 'changed' | 'draft';
    /**
     * The number of versions this Component has.
     *
     * @minimum 0
     * @TJS-type integer
     */
    versionCount?: number;
    /**
     * The number of versions that are currently staged.
     *
     * @minimum 0
     * @TJS-type integer
     */
    stagedVersionCount?: number;
    /**
     * The number of versions that are available to be embedded.
     *
     * @minimum 0
     * @TJS-type integer
     */
    embeddableVersionCount?: number;
    /**
     * The identifiers of the datasources for this Component.
     *
     * @items.type string
     */
    datasourceIds?: string[];
    /** The timestamp when the Component was created at */
    createdAt?: Date;
    /**
     * The timestamp when the Component was last staged at.
     *
     * @nullable
     */
    stagedAt?: Date;
    /**
     * The timestamp when the Component was last published at.
     *
     * @nullable
     */
    publishedAt?: Date;
    /** The timestamp when the Component was last modified at */
    modifiedAt?: Date;
    /** The User who last modified the Component */
    modifiedBy?: object;
}
export declare class Components extends SDKAdapter<Component> {
    fetch(component: Pick<ComponentParams, 'libraryId' | 'collectionId'>): Promise<Component[]>;
    get(component: Pick<ComponentParams, 'libraryId' | 'collectionId' | 'id'>): Promise<Component>;
    post(component: Atleast<ComponentParams, 'libraryId' | 'collectionId'>): Promise<Component>;
    put(component: Atleast<ComponentParams, 'libraryId' | 'collectionId' | 'id'>): Promise<Component>;
    delete(component: Pick<ComponentParams, 'libraryId' | 'collectionId' | 'id'>): Promise<Component>;
}
/**
 * This function returns default values for a component object in TypeScript.
 *
 * @param {SDK} sdk  - Instance of SDK (used for retrieving current user)
 * @returns The function `getComponentDefaults` is returning an object with default values for a component. These
 *          values will not be required when instantiating new instance of a Component model.
 */
export declare function getComponentDefaults(sdk: SDK): {
    id: string;
    status: string;
    createdAt: Date;
    modifiedAt: Date;
    versionCount: number;
    stagedVersionCount: number;
    embeddableVersionCount: number;
    datasourceIds: string[];
    description: string;
    modifiedBy: {
        id?: undefined;
        name?: undefined;
        givenName?: undefined;
        familyName?: undefined;
        email?: undefined;
        picture?: undefined;
        nickname?: undefined;
    } | {
        id: string;
        name: string;
        givenName: string;
        familyName: string;
        email: string;
        picture: string;
        nickname: string;
    };
};
export type ComponentMinimal<T extends keyof Component = never> = RequiredKeys<Component, typeof getComponentDefaults, T>;
/**
 * List of properties that component inherits from its collection. These become optional when instnatiating a new
 * component.
 */
export interface ComponentImplicit {
    library?: LibraryModel;
    libraryId: LibraryModel['id'];
    collectionId: CollectionModel['id'];
    collection?: CollectionModel;
}
export interface ComponentParams extends Component, ComponentImplicit {
}
export interface ComponentModel extends ComponentParams {
}
export declare class ComponentModel extends SDKModel<Component, ComponentMinimal, ComponentImplicit> implements ComponentParams {
    get adapter(): Components;
    builder: SDKModel;
    versions: ReturnType<this['constructVersionCollection']>;
    bundles: ReturnType<this['constructVersionBundleCollection']>;
    constructVersionCollection(): SDKCollection<typeof VersionModel, VersionModel, any, Omit<Pick<Partial<Version>, "model" | "id" | "status" | "createdAt" | "modifiedAt" | "revision" | "orderIdx" | "datasourceIds" | "modifiedBy" | "deletedAt" | "breakpointIds" | "classList" | "view" | "diff"> & Omit<Version, "model" | "id" | "status" | "createdAt" | "modifiedAt" | "revision" | "orderIdx" | "datasourceIds" | "modifiedBy" | "deletedAt" | "breakpointIds" | "classList" | "view" | "diff"> & Partial<import("./versions.js").VersionImplicit>, "collection" | "component" | "collectionId" | "libraryId" | "library" | "componentId">, {
        id: string;
    } | {
        details: {
            id: string;
        };
    } | {
        id: string;
        revision: number;
        status: "published" | "draft" | "staged" | "saved";
    }>;
    constructVersionBundleCollection(): SDKCollection<typeof import("./version-bundles.js").VersionBundleModel, import("./version-bundles.js").VersionBundleModel, any, Omit<Pick<Partial<import("./version-bundles.js").VersionBundle>, "id" | "createdAt" | "modifiedAt" | "breakpoints"> & Omit<import("./version-bundles.js").VersionBundle, "id" | "createdAt" | "modifiedAt" | "breakpoints"> & Partial<import("./version-bundles.js").VersionBundleImplicit>, "collection" | "component" | "collectionId" | "libraryId" | "library" | "componentId">, {
        id: string;
    } | {
        details: {
            id: string;
        };
    } | {
        id: string;
        revision: number;
        status: "published" | "draft" | "staged" | "saved";
    }>;
    getThumbnailURL(): string;
    getSourceURL(): string;
    getBundles(): ReturnType<this["constructVersionBundleCollection"]>;
    defineProperties(): void;
    getDefaults(sdk: SDK): {
        id: string;
        status: string;
        createdAt: Date;
        modifiedAt: Date;
        versionCount: number;
        stagedVersionCount: number;
        embeddableVersionCount: number;
        datasourceIds: string[];
        description: string;
        modifiedBy: {
            id?: undefined;
            name?: undefined;
            givenName?: undefined;
            familyName?: undefined;
            email?: undefined;
            picture?: undefined;
            nickname?: undefined;
        } | {
            id: string;
            name: string;
            givenName: string;
            familyName: string;
            email: string;
            picture: string;
            nickname: string;
        };
    };
    getHiddenProps(): string[];
    addVersion(values?: Partial<Version>): VersionModel;
    /**
     * Ensures that version with given `id` has given data. Best way to work with versioning, as it takes care about
     * drafts, changes and versions behind the scenes.
     */
    establishVersion(values?: Partial<Version>): VersionModel;
    changeVersions(callback: (versions: VersionModel[]) => any): Promise<void>;
    findVersion(id: VersionModel['id'], statuses?: VersionModel['status'][], revision?: VersionModel['revision']): VersionModel;
    /**
     * Returns a version based on container width provided. If more than one versions match width and versionName is
     * provided, returns version with specific name If more than one versions exist matching above criteria, it returns
     * the one having less breakpoints If no version matches width, returns the closest one
     */
    getVersionForWidth(width: number, versionName?: string, versions?: VersionModel[]): VersionModel;
    /**
     * This function returns the previous and next versions of a given version ID in an ordered list of versions.
     *
     * @param id  - The `id` parameter is of type `VersionModel['id']`, which means it is a specific property (`id`) of
     *            the `VersionModel` type. It is used as an identifier to find the neighbouring versions of a
     *            particular version.
     * @returns The function `getNeighbourVersions` returns an array with two elements. The first element is either the
     *          version object that comes before the version with the given `id` in the ordered list of versions, or
     *          the last version in the list if the given `id` is not found in the list. The second element is either
     *          the version object that comes after the version with the given `id`
     */
    getNeighbourVersions(id: VersionModel['id']): VersionModel[];
    getNotDeletedVersions(): VersionModel[];
    getVersions(status?: VersionModel['status'][] | VersionModel['status']): VersionModel[];
    /** Get versions that currently are staged and usable */
    getStagedVersions(): VersionModel[];
    /** Get versions that currently are published */
    getPublishedVersions(): VersionModel[];
    reduceVersionRevisions(versions?: VersionModel[]): import("./versions.js").VersionSnapshot[];
    /**
     * This function saves new and edited versions of a component.
     *
     * @param [autoSave=true]  - The `autoSave` parameter is a boolean value that determines whether or not to include
     *                         versions with a status of "saved" in the list of non-draft revisions to be saved. If
     *                         `autoSave` is `true`, then "saved" versions will be included. If `autoSave`. Default is
     *                         `true`
     * @returns A Promise that resolves when all the new and updated versions have been saved to the server.
     */
    saveVersions(autoSave?: boolean): Promise<VersionModel[]>;
    /**
     * This function aggregates version data for a component and returns an updated component object.
     *
     * @returns The updated component object after aggregating version data.
     */
    aggregateVersionData(versions?: ReturnType<this["constructVersionCollection"]>): this;
    /** Get settings for all datasoruces used in component */
    getDataSettings(): DataOptionsById | null;
    /** Get list of used datasources  */
    getDatasources(): DatasourceModel[];
    sortCompare(other: this): number;
    matchesSearch(search: string): boolean;
    getPath(): string;
    getBuilderPath(): string;
    static getNameWithoutIndex(name: string): string;
    static getIndexFromName(name: string): string;
    /**
     * Get all component versions that can be embedded (and thus need to be uploaded to CDN).
     * Turn bundles into versions as well.
     */
    getEmbeddableVersions(status: VersionModel['status']): VersionModel[];
    countEmbeddableVersions(versions?: VersionModel[]): number;
    getBundlableVersions(versions?: VersionModel[]): VersionModel[];
    static getAggregatedVersionData(allVersions: VersionModel[]): Pick<Component, "status" | "modifiedAt" | "versionCount" | "stagedVersionCount" | "embeddableVersionCount" | "datasourceIds" | "modifiedBy" | "stagedAt" | "publishedAt">;
    static getMaxIndex(component: ComponentModel, components: ComponentModel[]): number;
    /**
     * Creates a duplicate of a component with an incremented index in the name. Additionally ensures that all instanced
     * elements get unique id to avoid clashes. For that to work, it expects that versions are already fetched.
     *
     * @param component   - The component to duplicate.
     * @param components  - An array of existing components.
     * @returns A new component with the duplicated name and updated versions.
     */
    duplicate(component: ComponentModel, components: ComponentModel[]): this;
    static get schema(): SDKModelSchema<typeof ComponentModel, Component>;
    /** Copy component to a different collection (possibly from another library). */
    copyToCollection(collection: CollectionModel, options?: {
        versions: boolean;
        bundles: boolean;
    }): Promise<ComponentModel>;
    serialize(): {
        versions: Version[];
        bundles: import("./version-bundles.js").VersionBundle[];
        /** The identifier of the Library that Component belongs to */
        libraryId: string;
        /** The identifier of the Component */
        id: string;
        /** The identifier of the Collection */
        collectionId: string;
        /**
         * The name of the Component.
         *
         * @minLength 1
         * @maxLength 128
         */
        name: string;
        /**
         * The description of the Component.
         *
         * @minLength 0
         * @maxLength 128
         */
        description?: string;
        /** The status of the Component */
        status: "published" | "deleted" | "draft" | "staged" | "changed";
        /**
         * The number of versions this Component has.
         *
         * @minimum 0
         * @TJS-type integer
         */
        versionCount?: number;
        /**
         * The number of versions that are currently staged.
         *
         * @minimum 0
         * @TJS-type integer
         */
        stagedVersionCount?: number;
        /**
         * The number of versions that are available to be embedded.
         *
         * @minimum 0
         * @TJS-type integer
         */
        embeddableVersionCount?: number;
        /**
         * The identifiers of the datasources for this Component.
         *
         * @items.type string
         */
        datasourceIds?: string[];
        /** The timestamp when the Component was created at */
        createdAt?: Date;
        /**
         * The timestamp when the Component was last staged at.
         *
         * @nullable
         */
        stagedAt?: Date;
        /**
         * The timestamp when the Component was last published at.
         *
         * @nullable
         */
        publishedAt?: Date;
        /** The timestamp when the Component was last modified at */
        modifiedAt?: Date;
        /** The User who last modified the Component */
        modifiedBy?: object;
    };
}
//# sourceMappingURL=components.d.ts.map