import SDKAdapter from '../core/adapter.js';
import { SDKModel } from '../core/model.js';
import { SDKModelSchema } from '../core/schema.js';
import { SDKCollection } from '../index.js';
import type { SDK } from '../sdk.js';
import { Atleast, RequiredKeys } from '../utils/typescript.js';
import { LibraryModel, NanoId } from './libraries.js';
export type DataSourceId = string;
export type Data = Record<string, any> | any[];
export type DataScopes = Data | Record<DataSourceId, Data>;
export type DataOptionsById = Record<DataSourceId, DataSettings | Data>;
export type DataOptions = Data | DataSettings | DataOptionsById;
export type DataSettings = {
    params: {
        [key: string]: any;
    };
    headers: {
        [key: string]: string;
    };
    jsonpath: string;
    method: 'GET' | 'POST' | 'OPTIONS' | 'PUT' | 'DELETE' | 'UPDATE' | 'HEAD';
    body: {
        [key: string]: any;
    };
    url: string;
};
export declare function DataSettings(settings: any): DataSettings;
/** Defines the schema for Datasource. */
export interface Datasource {
    /** The identifier of the Library that Datasource belongs to */
    libraryId: NanoId;
    /** The identifier of the Datasource */
    id: NanoId;
    /**
     * The name of the Datasource.
     *
     * @minLength 1
     * @maxLength 128
     */
    name: string;
    /**
     * The description of the Datasource.
     *
     * @minLength 0
     * @maxLength 128
     */
    description: string;
    /** A sample of the Datasource */
    sample: object[] | object;
    /** The timestamp the Datascource was created at */
    createdAt?: Date;
    /** The timestamp when the Datascource was last modified at */
    modifiedAt?: Date;
    /**
     * The timestamp when the Datascource was last sampled at.
     *
     * @nullable
     */
    sampledAt?: Date;
    /**
     * Definition is a data returned from external resource, that we use to extract sample data from. Having this data, we
     * can detect if upstream structure was updated.
     *
     * @nullable
     */
    definition?: object | object[];
    /**
     * Definition of remote datasource for a conflicting datasource.
     *
     * @nullable
     */
    conflictDefinition?: object | object[];
    /**
     * Datasource type (internal, xm template, etc.)
     *
     * @nullable
     */
    type?: 'internal' | 'xmTemplate' | 'contentHubOne' | 'external';
    /**
     * Id of external source. Eg: Tenant name for XM Cloud datasources, Project Id for internal datasources etc.
     *
     * @nullable
     */
    externalSourceId?: string;
    /** Default settings for the datasource */
    settings: DataSettings;
}
export declare class Datasources extends SDKAdapter<Datasource> {
    fetch(datasource: Pick<DatasourceParams, 'libraryId'>): Promise<Datasource[]>;
    get(datasource: Pick<DatasourceParams, 'libraryId' | 'id'>): Promise<Datasource>;
    fetchUpdates(datasource: Pick<DatasourceParams, 'id' | 'libraryId' | 'settings'>): Promise<Datasource>;
    post(datasource: Atleast<DatasourceModel, 'libraryId'>): Promise<Datasource>;
    put(datasource: Atleast<DatasourceParams, 'libraryId' | 'id'>): Promise<Datasource>;
    delete(datasource: Pick<DatasourceParams, 'libraryId' | 'id'>): Promise<Datasource>;
}
export declare class XmDatasources extends Datasources {
    fetch(datasource: Pick<DatasourceParams, 'libraryId' | 'externalSourceId'>): Promise<Datasource[]>;
    fetchUpdates(datasource: Pick<DatasourceParams, 'id' | 'libraryId' | 'externalSourceId'>): Promise<Datasource>;
}
export declare class ContentHubOneDatasources extends Datasources {
    fetch(datasource: Pick<DatasourceParams, 'libraryId'>): Promise<Datasource[]>;
    fetchUpdates(datasource: Pick<DatasourceParams, 'id' | 'libraryId'>): Promise<Datasource>;
}
export declare function getDatasourceDefaults(sdk: SDK): {
    id: string;
    sampledAt: Date;
    createdAt: Date;
    modifiedAt: Date;
    type: string;
    definition: {};
    conflictDefinition: object;
    settings: DataSettings;
};
export type DatasourceMinimal<T extends keyof Datasource = never> = RequiredKeys<Datasource, typeof getDatasourceDefaults, T>;
export interface DatasourceImplicit {
    libraryId: LibraryModel['id'];
    library?: LibraryModel;
}
export interface DatasourceParams extends Datasource, DatasourceImplicit {
}
export interface DatasourceModel extends DatasourceParams {
}
export declare class DatasourceModel extends SDKModel<Datasource, DatasourceMinimal, DatasourceImplicit> implements DatasourceParams {
    get adapter(): Datasources | XmDatasources;
    fetchUpdates(): Promise<this>;
    getCollection(): SDKCollection<typeof DatasourceModel, DatasourceModel, Datasource, Omit<Pick<Partial<Datasource>, "type" | "id" | "definition" | "settings" | "createdAt" | "modifiedAt" | "sampledAt" | "conflictDefinition"> & Omit<Datasource, "type" | "id" | "definition" | "settings" | "createdAt" | "modifiedAt" | "sampledAt" | "conflictDefinition"> & Partial<DatasourceImplicit>, "libraryId" | "externalSourceId" | "library">, {
        id: string;
    } | {
        details: {
            id: string;
        };
    } | {
        id: string;
        revision: number;
        status: "published" | "draft" | "staged" | "saved";
    }> | SDKCollection<typeof DatasourceModel, DatasourceModel, Datasource, Omit<Pick<Partial<Datasource>, "type" | "id" | "definition" | "settings" | "createdAt" | "modifiedAt" | "sampledAt" | "conflictDefinition"> & Omit<Datasource, "type" | "id" | "definition" | "settings" | "createdAt" | "modifiedAt" | "sampledAt" | "conflictDefinition"> & Partial<DatasourceImplicit>, "type" | "libraryId" | "library">, {
        id: string;
    } | {
        details: {
            id: string;
        };
    } | {
        id: string;
        revision: number;
        status: "published" | "draft" | "staged" | "saved";
    }> | SDKCollection<typeof DatasourceModel, DatasourceModel, Datasource, Omit<Pick<Partial<Datasource>, "type" | "id" | "definition" | "settings" | "createdAt" | "modifiedAt" | "sampledAt" | "conflictDefinition"> & Omit<Datasource, "type" | "id" | "definition" | "settings" | "createdAt" | "modifiedAt" | "sampledAt" | "conflictDefinition"> & Partial<DatasourceImplicit>, "type" | "externalSourceId">, {
        id: string;
    } | {
        details: {
            id: string;
        };
    } | {
        id: string;
        revision: number;
        status: "published" | "draft" | "staged" | "saved";
    }>;
    getHiddenProps(): string[];
    getDefaults(sdk: SDK): {
        id: string;
        sampledAt: Date;
        createdAt: Date;
        modifiedAt: Date;
        type: string;
        definition: {};
        conflictDefinition: object;
        settings: DataSettings;
    };
    isInternal(): boolean;
    isExtended(): boolean;
    static loadFromSessionStorage(): Datasource[];
    static storeToSessionStorage(datasources: Datasource[]): void;
    static registeredDatasources: Datasource[];
    static setExternalDatasources(datasources: Datasource[]): void;
    static updateExternalDatasources(datasources: SDKCollection<typeof DatasourceModel>, list?: Datasource[]): void;
    getPath(): string;
    isConflictingTowards(previousState: Datasource): boolean;
    /**
     * Looks for changes in the generated sample of the definition of a datasource. It only looks for removal of fields
     * though. Addition of fields is not something that concerns us
     */
    static isConflicting(currentSample: Datasource['sample'], previousSample: Datasource['sample']): boolean;
    static getTreeViewLeafState(previous: any, current: any): "new" | "removed" | "changed" | "unchanged";
    static getSettingsFromDatasources(datasourceIds: string[], datasources: DatasourceModel[]): Record<string, DataSettings>;
    static getExternalComponentDatasources(data: DataOptionsById, allDatasources: DatasourceModel[]): DatasourceModel[];
    static get schema(): SDKModelSchema<typeof DatasourceModel, Datasource>;
    copyToLibrary(library: LibraryModel): Promise<DatasourceModel>;
}
/**
 * Converts settings object into fetch function arguments.
 *
 * @param {any} settings  - The settings object.
 * @returns {Array} - An array containing URL and options objects as fetch function arguments.
 */
export declare function toFetchArguments(settings?: any): Parameters<typeof fetch>;
//# sourceMappingURL=datasources.d.ts.map