import SDKAdapter from '../core/adapter.js';
import { SDKCollection } from '../core/collection.js';
import { SDKModel } from '../core/model.js';
import { SDKModelSchema } from '../core/schema.js';
import type { SDK } from '../sdk.js';
import { Atleast, RequiredKeys } from '../utils/typescript.js';
import { StylesheetModel } from './stylesheets.js';
export declare class LibraryError extends Error {
    name: string;
}
/**
 * The NanoId identifier.
 *
 * @minLength 3
 * @maxLength 64
 */
export type NanoId = string;
/** Defines the schema for Library. */
export interface Library {
    /** The identifier of the Library */
    id: NanoId;
    /**
     * The name of the Library.
     *
     * @minLength 1
     * @maxLength 128
     */
    name: string;
    /**
     * The identifier referencing XM project library is created for. This enables additional security features that only
     * allow XM Users having access to that project access the library.
     *
     * @minLength 1
     * @maxLength 128
     */
    xmProjectId?: string;
    /**
     * API Key is used to access the library programmatically.
     *
     * @minLength 0
     * @maxLength 64
     */
    apiKey?: string;
    organizationId?: string;
    /**
     * ID of a library that should be used as a template on creation.
     *
     * @minLength 0
     * @maxLength 64
     * @nullable
     */
    templateId?: string | null;
    /** Library settings */
    settings?: LibrarySettings;
}
export declare class Libraries extends SDKAdapter<Library> {
    get({ id, systemId }: Pick<Library, 'id'> & {
        systemId: string;
    }): Promise<any>;
    fetch({ organizationId }: Pick<Library, 'organizationId'>): Promise<Library[]>;
    post(library: Atleast<LibraryParams, 'id' | 'name'>): Promise<Library>;
    put(library: Partial<LibraryParams>): Promise<Library>;
}
export declare class TemplateLibraries extends SDKAdapter<Library> {
    fetch({ organizationId }: Pick<Library, 'organizationId'>): Promise<Library[]>;
}
export declare function getLibraryDefaults(sdk: SDK): {
    id: string;
    apiKey: string;
    createdAt: Date;
    modifiedAt: Date;
    templateId: string;
    settings: {};
};
export type LibraryMinimal<T extends keyof Library = never> = RequiredKeys<Library, typeof getLibraryDefaults, T>;
export interface LibraryParams extends Library {
}
export interface LibraryModel extends LibraryParams {
}
export interface LibrarySettings {
    externalPreviewURL?: string;
    renderingHostEnv?: string;
    templates?: string[];
    contentHubOneDeliveryKey?: string;
    contentHubOnePreviewKey?: string;
}
export declare class LibraryModel extends SDKModel<Library, LibraryMinimal> implements LibraryParams {
    collections: ReturnType<this['constructCollections']>;
    datasources: ReturnType<this['constructDatasources']>;
    contentHubOneDatasources: ReturnType<this['constructContentHubOneDatasources']>;
    stylesheets: ReturnType<this['constructStylesheets']>;
    components: ReturnType<this['constructComponents']>;
    settings: LibrarySettings;
    get adapter(): Libraries;
    defineProperties(): void;
    getHiddenProps(): string[];
    constructCollections(): SDKCollection<typeof import("./collections.js").CollectionModel, import("./collections.js").CollectionModel, import("./collections.js").Collection, Omit<Pick<Partial<import("./collections.js").Collection>, "id" | "isDefault" | "createdAt" | "modifiedAt" | "orderIdx"> & Omit<import("./collections.js").Collection, "id" | "isDefault" | "createdAt" | "modifiedAt" | "orderIdx"> & Partial<import("./collections.js").CollectionImplicit>, "libraryId" | "library">, {
        id: string;
    } | {
        details: {
            id: string;
        };
    } | {
        id: string;
        revision: number;
        status: "published" | "draft" | "staged" | "saved";
    }>;
    constructComponents(): SDKCollection<typeof import("./components.js").ComponentModel, import("./components.js").ComponentModel, import("./components.js").Component, Omit<Pick<Partial<import("./components.js").Component>, "description" | "id" | "status" | "createdAt" | "modifiedAt" | "versionCount" | "stagedVersionCount" | "embeddableVersionCount" | "datasourceIds" | "modifiedBy"> & Omit<import("./components.js").Component, "description" | "id" | "status" | "createdAt" | "modifiedAt" | "versionCount" | "stagedVersionCount" | "embeddableVersionCount" | "datasourceIds" | "modifiedBy"> & Partial<import("./components.js").ComponentImplicit>, "libraryId" | "library">, {
        id: string;
    } | {
        details: {
            id: string;
        };
    } | {
        id: string;
        revision: number;
        status: "published" | "draft" | "staged" | "saved";
    }>;
    constructDatasources(): SDKCollection<typeof import("./datasources.js").DatasourceModel, import("./datasources.js").DatasourceModel, import("./datasources.js").Datasource, Omit<Pick<Partial<import("./datasources.js").Datasource>, "type" | "id" | "definition" | "settings" | "createdAt" | "modifiedAt" | "sampledAt" | "conflictDefinition"> & Omit<import("./datasources.js").Datasource, "type" | "id" | "definition" | "settings" | "createdAt" | "modifiedAt" | "sampledAt" | "conflictDefinition"> & Partial<import("./datasources.js").DatasourceImplicit>, "libraryId" | "externalSourceId" | "library">, {
        id: string;
    } | {
        details: {
            id: string;
        };
    } | {
        id: string;
        revision: number;
        status: "published" | "draft" | "staged" | "saved";
    }>;
    constructContentHubOneDatasources(): SDKCollection<typeof import("./datasources.js").DatasourceModel, import("./datasources.js").DatasourceModel, import("./datasources.js").Datasource, Omit<Pick<Partial<import("./datasources.js").Datasource>, "type" | "id" | "definition" | "settings" | "createdAt" | "modifiedAt" | "sampledAt" | "conflictDefinition"> & Omit<import("./datasources.js").Datasource, "type" | "id" | "definition" | "settings" | "createdAt" | "modifiedAt" | "sampledAt" | "conflictDefinition"> & Partial<import("./datasources.js").DatasourceImplicit>, "type" | "libraryId" | "library">, {
        id: string;
    } | {
        details: {
            id: string;
        };
    } | {
        id: string;
        revision: number;
        status: "published" | "draft" | "staged" | "saved";
    }>;
    constructStylesheets(): SDKCollection<typeof StylesheetModel, StylesheetModel, import("./stylesheets.js").Stylesheet, Omit<Pick<Partial<import("./stylesheets.js").Stylesheet>, "source" | "id" | "status" | "createdAt" | "modifiedAt" | "formatVersion" | "revision"> & Omit<import("./stylesheets.js").Stylesheet, "source" | "id" | "status" | "createdAt" | "modifiedAt" | "formatVersion" | "revision"> & Partial<import("./stylesheets.js").StylesheetImplicit>, "libraryId" | "library">, {
        id: string;
    } | {
        details: {
            id: string;
        };
    } | {
        id: string;
        revision: number;
        status: "published" | "draft" | "staged" | "saved";
    }>;
    get stylesheet(): StylesheetModel;
    set stylesheet(stylesheet: StylesheetModel);
    getDefaults(sdk: SDK): {
        id: string;
        apiKey: string;
        createdAt: Date;
        modifiedAt: Date;
        templateId: string;
        settings: {};
    };
    getPath(): string;
    fetchAll(includeExternalDatasources?: boolean, includeVersions?: boolean): Promise<(SDKCollection<typeof import("./datasources.js").DatasourceModel, import("./datasources.js").DatasourceModel, import("./datasources.js").Datasource, Omit<Pick<Partial<import("./datasources.js").Datasource>, "type" | "id" | "definition" | "settings" | "createdAt" | "modifiedAt" | "sampledAt" | "conflictDefinition"> & Omit<import("./datasources.js").Datasource, "type" | "id" | "definition" | "settings" | "createdAt" | "modifiedAt" | "sampledAt" | "conflictDefinition"> & Partial<import("./datasources.js").DatasourceImplicit>, "type" | "externalSourceId">, {
        id: string;
    } | {
        details: {
            id: string;
        };
    } | {
        id: string;
        revision: number;
        status: "published" | "draft" | "staged" | "saved";
    }> | (SDKCollection<typeof import("./versions.js").VersionModel, import("./versions.js").VersionModel, any, Omit<Pick<Partial<import("./versions.js").Version>, "model" | "id" | "status" | "createdAt" | "modifiedAt" | "revision" | "orderIdx" | "datasourceIds" | "modifiedBy" | "deletedAt" | "breakpointIds" | "classList" | "view" | "diff"> & Omit<import("./versions.js").Version, "model" | "id" | "status" | "createdAt" | "modifiedAt" | "revision" | "orderIdx" | "datasourceIds" | "modifiedBy" | "deletedAt" | "breakpointIds" | "classList" | "view" | "diff"> & Partial<import("./versions.js").VersionImplicit>, "collection" | "component" | "collectionId" | "libraryId" | "library" | "componentId">, {
        id: string;
    } | {
        details: {
            id: string;
        };
    } | {
        id: string;
        revision: number;
        status: "published" | "draft" | "staged" | "saved";
    }> | Promise<SDKCollection<typeof import("./versions.js").VersionModel, import("./versions.js").VersionModel, any, Omit<Pick<Partial<import("./versions.js").Version>, "model" | "id" | "status" | "createdAt" | "modifiedAt" | "revision" | "orderIdx" | "datasourceIds" | "modifiedBy" | "deletedAt" | "breakpointIds" | "classList" | "view" | "diff"> & Omit<import("./versions.js").Version, "model" | "id" | "status" | "createdAt" | "modifiedAt" | "revision" | "orderIdx" | "datasourceIds" | "modifiedBy" | "deletedAt" | "breakpointIds" | "classList" | "view" | "diff"> & Partial<import("./versions.js").VersionImplicit>, "collection" | "component" | "collectionId" | "libraryId" | "library" | "componentId">, {
        id: string;
    } | {
        details: {
            id: string;
        };
    } | {
        id: string;
        revision: number;
        status: "published" | "draft" | "staged" | "saved";
    }>>[])[] | Awaited<ReturnType<this["constructStylesheets"]>> | Awaited<ReturnType<this["constructDatasources"]>> | Awaited<ReturnType<this["constructContentHubOneDatasources"]>>)[]>;
    serialize(): {
        collections: {
            components: {
                versions: import("./versions.js").Version[];
                bundles: import("./version-bundles.js").VersionBundle[];
                libraryId: string;
                id: string;
                collectionId: string;
                name: string;
                description?: string;
                status: "published" | "deleted" | "draft" | "staged" | "changed";
                versionCount?: number;
                stagedVersionCount?: number;
                embeddableVersionCount?: number;
                datasourceIds?: string[];
                createdAt?: Date;
                stagedAt?: Date;
                publishedAt?: Date;
                modifiedAt?: Date;
                modifiedBy?: object;
            }[];
            libraryId: string;
            id: string;
            name: string;
            /**
             * The name of the Library.
             *
             * @minLength 1
             * @maxLength 128
             */
            isDefault: boolean;
            orderIdx?: string;
            createdAt?: Date;
            modifiedAt?: Date;
        }[];
        stylesheets: import("./stylesheets.js").Stylesheet[];
        datasources: import("./datasources.js").Datasource[];
        /** The identifier of the Library */
        id: string;
        /**
         * The name of the Library.
         *
         * @minLength 1
         * @maxLength 128
         */
        name: string;
        /**
         * The identifier referencing XM project library is created for. This enables additional security features that only
         * allow XM Users having access to that project access the library.
         *
         * @minLength 1
         * @maxLength 128
         */
        xmProjectId?: string;
        /**
         * API Key is used to access the library programmatically.
         *
         * @minLength 0
         * @maxLength 64
         */
        apiKey?: string;
        organizationId?: string;
        /**
         * ID of a library that should be used as a template on creation.
         *
         * @minLength 0
         * @maxLength 64
         * @nullable
         */
        templateId?: string;
        /** Library settings */
        settings?: LibrarySettings;
    };
    setXMProjectData({ authorization, systemId }: {
        authorization?: string;
        systemId: string;
    }): Promise<void>;
    getPreviewURL(): string;
    getFallbackPreviewURL(): string;
    decorate(navigate: (to: string, options: {
        replace?: boolean;
        state?: any;
    }) => void): void;
    static get schema(): SDKModelSchema<typeof LibraryModel, Library>;
    onDatasourceChange(): void;
}
//# sourceMappingURL=libraries.d.ts.map