import SDKAdapter from '../core/adapter.js';
import { SDKCollection, SDKFunctionalCollection } from '../core/collection.js';
import { SDKModel } from '../core/model.js';
import { SDKModelSchema } from '../core/schema.js';
import type { SDK } from '../sdk.js';
import * as Style from '../style/index.js';
import { Atleast, RequiredKeys } from '../utils/typescript.js';
import type { LibraryModel, NanoId } from './libraries.js';
/** Defines the schema for Stylesheet. */
export interface Stylesheet {
    /** The identifier of the Library that Stylesheet belongs to */
    libraryId: NanoId;
    /** The identifier of the Stylesheet */
    id: NanoId;
    /** The source of the Stylesheet */
    source: any;
    /**
     * The format version of the Stylesheet.
     *
     * @minimum 0
     * @TJS-type integer
     */
    formatVersion: number;
    /** The status of the Stylesheet */
    status: 'published' | 'draft' | 'staged' | 'saved';
    /**
     * The numeric revision of the Stylesheet.
     *
     * @minimum 0
     * @TJS-type integer
     */
    revision: number;
    /** The timestamp the Stylesheet was created at */
    createdAt: Date;
    /** The timestamp when the Stylesheet was last modified at */
    modifiedAt: Date;
}
export declare function getStylesheetDefaults(sdk: SDK): {
    id: string;
    modifiedAt: Date;
    createdAt: Date;
    formatVersion: number;
    source: (Style.RuleWithProps<"fill", {
        backgroundImage: any;
        backgroundPositionX: Style.Length;
        backgroundPositionY: Style.Length;
        backgroundRepeat: Style.TypeDefinitions.fill.BackgroundRepeat;
        backgroundSize: Style.TypeDefinitions.fill.BackgroundSize;
        blur: Style.Length;
        opacity: number;
        backgroundColor: Style.ColorReference;
        backgroundGradient: {
            angle: number;
            stops: {
                color: Style.ColorReference | {
                    red: number;
                    green: number;
                    blue: number;
                    alpha: number;
                };
                start: Style.Length;
                end: Style.Length;
            }[];
        };
    }> | Style.RuleWithProps<"color", {
        red: number;
        green: number;
        blue: number;
        alpha: number;
    }> | Style.RuleWithProps<"breakpoint", {
        minWidth: number;
        maxWidth: number;
        query: string;
    }> | Style.RuleWithProps<"collection", {
        type: "fill" | "color" | "breakpoint" | "collection" | "font" | "graphic" | "decoration" | "dimensions" | "grid" | "layout" | "lines" | "media" | "palette" | "spacing" | "typography" | "block" | "inline" | "text" | "theme";
    }> | Style.RuleWithProps<"graphic", {
        url: any;
    }> | Style.RuleWithProps<"decoration", {
        boxShadowInset: boolean;
        boxShadowOffsetX: Style.Length;
        boxShadowOffsetY: Style.Length;
        boxShadowBlurRadius: Style.Length;
        boxShadowSpreadRadius: Style.Length;
        boxShadowColor: Style.ColorReference;
        borderTopLeftRadius: Style.Length;
        borderTopRightRadius: Style.Length;
        borderBottomLeftRadius: Style.Length;
        borderBottomRightRadius: Style.Length;
        borderStyle: Style.TypeDefinitions.decoration.BorderStyle;
        borderColor: Style.ColorReference;
        borderLeftWidth: Style.Length;
        borderRightWidth: Style.Length;
        borderTopWidth: Style.Length;
        borderBottomWidth: Style.Length;
    }> | Style.RuleWithProps<"grid", {
        columns: {
            min: Style.Length;
            max: Style.Length;
        }[];
        rows: {
            min: Style.Length;
            max: Style.Length;
        }[];
        items: {
            order: number;
            from: {
                row: number;
                column: number;
            };
            to: {
                row: number;
                column: number;
            };
        }[];
    }> | Style.RuleWithProps<"layout", {
        flexWrap: boolean;
        columnCount: number;
        weights: number[];
        justifyContent: Style.TypeDefinitions.layout.JustifyContent;
        alignItems: Style.TypeDefinitions.layout.AlignItems;
    }> | Style.RuleWithProps<"palette", {
        textColor: Style.ColorReference;
        linkDecoration: Style.TypeDefinitions.palette.LinkDecorationStyle;
    }> | Style.RuleWithProps<"spacing", {
        paddingTop: Style.Length;
        paddingBottom: Style.Length;
        paddingLeft: Style.Length;
        paddingRight: Style.Length;
        rowGap: Style.Length;
        columnGap: Style.Length;
    }> | Style.RuleWithProps<"typography", {
        fontVariant: any;
        allowBoldAndItalic: boolean;
        textTransform: any;
        base: {
            fontSize: Style.Length;
            lineHeight: Style.Length;
            letterSpacing: Style.Length;
            paragraphSpacing: Style.Length;
            iconSize: Style.Length;
        };
        overrides: Record<string, {
            fontSize: Style.Length;
            lineHeight: Style.Length;
            letterSpacing: Style.Length;
            paragraphSpacing: Style.Length;
            iconSize: Style.Length;
        }>;
    }> | Style.RuleWithProps<"theme", {
        blocks: {
            id: string;
            refId: string;
            fillId: string;
            decorationId: string;
            spacingId: string;
            typographyId: string;
            paletteId: string;
            isDefault: boolean;
        }[];
        inlines: {
            id: string;
            refId: string;
            fillId: string;
            decorationId: string;
            spacingId: string;
            typographyId: string;
            paletteId: string;
            isDefault: boolean;
        }[];
        texts: {
            id: string;
            refId: string;
            fillId: string;
            decorationId: string;
            spacingId: string;
            typographyId: string;
            paletteId: string;
            isDefault: boolean;
        }[];
    }>)[];
    revision: number;
    status: string;
};
export declare class Stylesheets extends SDKAdapter<Stylesheet> {
    fetch(stylesheet: Pick<StylesheetParams, 'libraryId'>): Promise<Stylesheet[]>;
    get(stylesheet: Pick<StylesheetParams, 'libraryId' | 'status'>): Promise<Stylesheet>;
    post(stylesheet: Atleast<StylesheetParams, 'libraryId'>): Promise<Stylesheet>;
    put(stylesheet: Atleast<StylesheetParams, 'libraryId'>): Promise<Stylesheet>;
    prune({ libraryId }?: Partial<Pick<Stylesheet, 'libraryId'>>): void;
}
export interface StylesheetImplicit {
    libraryId: LibraryModel['id'];
    library?: LibraryModel;
}
export type StylesheetMinimal<T extends keyof Stylesheet = never> = RequiredKeys<Stylesheet, typeof getStylesheetDefaults, T>;
export interface StylesheetParams extends Stylesheet, StylesheetImplicit {
}
export interface StylesheetModel extends StylesheetParams {
}
export declare class StylesheetModel extends SDKModel<Stylesheet, StylesheetMinimal, StylesheetImplicit> implements StylesheetParams {
    _source: Style.Rule[];
    ['constructor']: typeof StylesheetModel;
    get adapter(): Stylesheets;
    rules: ReturnType<this['constructRules']>;
    static getCSS(rules: Style.Rule[]): string;
    _css: string;
    get css(): string;
    set css(css: string);
    getId(): string;
    getSourcePath(): string;
    defineProperties(): void;
    /** Rules is an observable collection that combines `source` styles + `internal` styles. Only source styles are saved on backend. */
    constructRules(): SDKFunctionalCollection<Style.Rule<"fill" | "color" | "breakpoint" | "collection" | "font" | "graphic" | "decoration" | "dimensions" | "grid" | "layout" | "lines" | "media" | "palette" | "spacing" | "typography" | "block" | "inline" | "text" | "theme">, Pick<Style.Rule<"fill" | "color" | "breakpoint" | "collection" | "font" | "graphic" | "decoration" | "dimensions" | "grid" | "layout" | "lines" | "media" | "palette" | "spacing" | "typography" | "block" | "inline" | "text" | "theme">, "type">>;
    sortStyles(rules?: Style.Rule[]): Style.Rule<"fill" | "color" | "breakpoint" | "collection" | "font" | "graphic" | "decoration" | "dimensions" | "grid" | "layout" | "lines" | "media" | "palette" | "spacing" | "typography" | "block" | "inline" | "text" | "theme">[];
    getHiddenProps(): string[];
    getDefaults(sdk: SDK): {
        id: string;
        modifiedAt: Date;
        createdAt: Date;
        formatVersion: number;
        source: (Style.RuleWithProps<"fill", {
            backgroundImage: any;
            backgroundPositionX: Style.Length;
            backgroundPositionY: Style.Length;
            backgroundRepeat: Style.TypeDefinitions.fill.BackgroundRepeat;
            backgroundSize: Style.TypeDefinitions.fill.BackgroundSize;
            blur: Style.Length;
            opacity: number;
            backgroundColor: Style.ColorReference;
            backgroundGradient: {
                angle: number;
                stops: {
                    color: Style.ColorReference | {
                        red: number;
                        green: number;
                        blue: number;
                        alpha: number;
                    };
                    start: Style.Length;
                    end: Style.Length;
                }[];
            };
        }> | Style.RuleWithProps<"color", {
            red: number;
            green: number;
            blue: number;
            alpha: number;
        }> | Style.RuleWithProps<"breakpoint", {
            minWidth: number;
            maxWidth: number;
            query: string;
        }> | Style.RuleWithProps<"collection", {
            type: "fill" | "color" | "breakpoint" | "collection" | "font" | "graphic" | "decoration" | "dimensions" | "grid" | "layout" | "lines" | "media" | "palette" | "spacing" | "typography" | "block" | "inline" | "text" | "theme";
        }> | Style.RuleWithProps<"graphic", {
            url: any;
        }> | Style.RuleWithProps<"decoration", {
            boxShadowInset: boolean;
            boxShadowOffsetX: Style.Length;
            boxShadowOffsetY: Style.Length;
            boxShadowBlurRadius: Style.Length;
            boxShadowSpreadRadius: Style.Length;
            boxShadowColor: Style.ColorReference;
            borderTopLeftRadius: Style.Length;
            borderTopRightRadius: Style.Length;
            borderBottomLeftRadius: Style.Length;
            borderBottomRightRadius: Style.Length;
            borderStyle: Style.TypeDefinitions.decoration.BorderStyle;
            borderColor: Style.ColorReference;
            borderLeftWidth: Style.Length;
            borderRightWidth: Style.Length;
            borderTopWidth: Style.Length;
            borderBottomWidth: Style.Length;
        }> | Style.RuleWithProps<"grid", {
            columns: {
                min: Style.Length;
                max: Style.Length;
            }[];
            rows: {
                min: Style.Length;
                max: Style.Length;
            }[];
            items: {
                order: number;
                from: {
                    row: number;
                    column: number;
                };
                to: {
                    row: number;
                    column: number;
                };
            }[];
        }> | Style.RuleWithProps<"layout", {
            flexWrap: boolean;
            columnCount: number;
            weights: number[];
            justifyContent: Style.TypeDefinitions.layout.JustifyContent;
            alignItems: Style.TypeDefinitions.layout.AlignItems;
        }> | Style.RuleWithProps<"palette", {
            textColor: Style.ColorReference;
            linkDecoration: Style.TypeDefinitions.palette.LinkDecorationStyle;
        }> | Style.RuleWithProps<"spacing", {
            paddingTop: Style.Length;
            paddingBottom: Style.Length;
            paddingLeft: Style.Length;
            paddingRight: Style.Length;
            rowGap: Style.Length;
            columnGap: Style.Length;
        }> | Style.RuleWithProps<"typography", {
            fontVariant: any;
            allowBoldAndItalic: boolean;
            textTransform: any;
            base: {
                fontSize: Style.Length;
                lineHeight: Style.Length;
                letterSpacing: Style.Length;
                paragraphSpacing: Style.Length;
                iconSize: Style.Length;
            };
            overrides: Record<string, {
                fontSize: Style.Length;
                lineHeight: Style.Length;
                letterSpacing: Style.Length;
                paragraphSpacing: Style.Length;
                iconSize: Style.Length;
            }>;
        }> | Style.RuleWithProps<"theme", {
            blocks: {
                id: string;
                refId: string;
                fillId: string;
                decorationId: string;
                spacingId: string;
                typographyId: string;
                paletteId: string;
                isDefault: boolean;
            }[];
            inlines: {
                id: string;
                refId: string;
                fillId: string;
                decorationId: string;
                spacingId: string;
                typographyId: string;
                paletteId: string;
                isDefault: boolean;
            }[];
            texts: {
                id: string;
                refId: string;
                fillId: string;
                decorationId: string;
                spacingId: string;
                typographyId: string;
                paletteId: string;
                isDefault: boolean;
            }[];
        }>)[];
        revision: number;
        status: string;
    };
    getCollectionItems(style: Style.Rule): Style.Rule<"fill" | "color" | "breakpoint" | "collection" | "font" | "graphic" | "decoration" | "dimensions" | "grid" | "layout" | "lines" | "media" | "palette" | "spacing" | "typography" | "block" | "inline" | "text" | "theme">[];
    getStatus(): Stylesheet['status'];
    getBreakpointForWidth(width: number, breakpoints?: Style.RuleWithProps<"breakpoint", {
        minWidth: number;
        maxWidth: number;
        query: string;
    }>[]): Style.RuleWithProps<"breakpoint", {
        minWidth: number;
        maxWidth: number;
        query: string;
    }>;
    isUpToDateWith(status: Stylesheet['status']): boolean;
    isSameContent(stylesheet: StylesheetModel): boolean;
    /** Get highest revision of a stylesheet matching current status(es) if given */
    getCurrent(statuses?: Stylesheet['status'][] | Stylesheet['status']): StylesheetModel;
    getPath(): string;
    /** Return a current draft or create a new draft bumping up the revision */
    getDraft(): StylesheetModel;
    getComparedStylesheet(status: Stylesheet['status']): StylesheetModel;
    /** Register a new revision of a stylesheet with given properties based on this stylesheet  */
    produceNewVersion(props?: Partial<Stylesheet>): StylesheetModel;
    /** Delete rule from a stylesheet, creates draft */
    deleteRule(rule: Style.Rule): void;
    /** Update or insert a rule in a stylesheet, creates draft */
    upsertRule(rule: Style.Rule): Style.Rule<"fill" | "color" | "breakpoint" | "collection" | "font" | "graphic" | "decoration" | "dimensions" | "grid" | "layout" | "lines" | "media" | "palette" | "spacing" | "typography" | "block" | "inline" | "text" | "theme">;
    /** Get last revision of a stylesheet that was ever saved */
    getLastSaved(): StylesheetModel;
    /** Save draft if it has changes over currently saved stylesheet */
    saveDraft(): Promise<void>;
    stage(): Promise<StylesheetModel>;
    publish(): Promise<StylesheetModel>;
    revertDraftTo(): Promise<StylesheetModel>;
    /** Get draft if it has changes over currently saved version, will ignore meaningless drafts */
    getChangedDraft(): StylesheetModel;
    /** Filter out all internal styles that dont need to be saved on backend */
    getSourceRules(rules?: Style.Rule[]): Style.Rule<"fill" | "color" | "breakpoint" | "collection" | "font" | "graphic" | "decoration" | "dimensions" | "grid" | "layout" | "lines" | "media" | "palette" | "spacing" | "typography" | "block" | "inline" | "text" | "theme">[];
    /** Copy stylesheet over to the other library. It will maintain revision numbering if the library already has stylesheet. Note that it's important that stylesheet is first saved and then added to collection, otherwise it will not be able to handle concurrent copyToLibrary calls and assign revisions correctly */
    copyToLibrary(target: LibraryModel, forceStatus?: StylesheetModel['status']): Promise<StylesheetModel>;
    get source(): Style.Rule[];
    set source(rules: Style.Rule[]);
    static get version(): number;
    static get schema(): SDKModelSchema<typeof StylesheetModel, Stylesheet>;
    static getStyleNameWithoutIndex(name: string): string;
    static getIndexFromStyleName(name: string): string;
    static getStyleMaxIndex(style: Style.Rule, styles: Style.Rule[]): number;
    static generateNewComboIds(style: any): any;
    static statuses: string[];
    static sort(revisions: StylesheetModel[]): StylesheetModel[];
    static onCollectionChange(versions: SDKCollection<typeof StylesheetModel>): StylesheetModel[];
    duplicateRule(rule: Style.Rule): Style.Rule<"fill" | "color" | "breakpoint" | "collection" | "font" | "graphic" | "decoration" | "dimensions" | "grid" | "layout" | "lines" | "media" | "palette" | "spacing" | "typography" | "block" | "inline" | "text" | "theme">;
    static getDuplicateRule(rule: Style.Rule, rules: Style.Rule[]): Style.Rule<"fill" | "color" | "breakpoint" | "collection" | "font" | "graphic" | "decoration" | "dimensions" | "grid" | "layout" | "lines" | "media" | "palette" | "spacing" | "typography" | "block" | "inline" | "text" | "theme">[];
    static getStyleTabSummary<S extends Style.Rule.ElementAspect, F extends Style.Form>(style: S, form: F): string;
    getRuleErrors<T extends Style.Rule = Style.Rule>(rule: T, rules?: Style.Rule[]): Style.RuleError<T>;
    get Style(): typeof Style;
    get ClassList(): typeof Style.ClassList;
    get Context(): typeof Style.ContextMethods & {
        getClassList(context: Style.ContextOrClassList, ...extraClasses: string[]): string[];
        getDefinition(context: Style.ContextOrClassList): {
            readonly name: "paragraph";
            readonly id: "paragraph";
            readonly type: "text";
            readonly htmlTagName: "p";
            readonly className: "-paragraph";
            readonly disguise: string;
            readonly isImplicit: true;
            readonly isCustom: false;
            readonly placeholder: "Text element";
            readonly title: "Paragraphs";
            readonly label: "Paragraph";
            readonly description: "Regular text element used for all text and copy that isn't heading";
        } | {
            readonly name: "heading1";
            readonly id: "heading1";
            readonly type: "text";
            readonly htmlTagName: "h1";
            readonly className: "-heading1";
            readonly disguise: string;
            readonly isImplicit: true;
            readonly isCustom: false;
            readonly title: "Heading 1";
            readonly label: "Heading 1";
            readonly description: "Main heading of the document, used for largest of headings";
        } | {
            readonly name: "heading2";
            readonly id: "heading2";
            readonly type: "text";
            readonly htmlTagName: "h2";
            /** The timestamp the Stylesheet was created at */
            readonly className: "-heading2";
            readonly disguise: string;
            readonly isImplicit: true;
            readonly isCustom: false;
            readonly title: "Heading 2";
            readonly label: "Heading 2";
            readonly description: "Second level of headings";
        } | {
            readonly name: "heading3";
            readonly id: "heading3";
            readonly type: "text";
            readonly htmlTagName: "h3";
            readonly className: "-heading3";
            readonly disguise: string;
            readonly isImplicit: true;
            readonly isCustom: false;
            readonly title: "Heading 3";
            readonly label: "Heading 3";
            readonly description: "Third level of headings often used for taglines";
        } | {
            readonly name: "heading4";
            readonly id: "heading4";
            readonly type: "text";
            readonly htmlTagName: "h4";
            readonly className: "-heading4";
            readonly disguise: string;
            readonly isImplicit: true;
            readonly isCustom: false;
            readonly title: "Heading 4";
            readonly label: "Heading 4";
            readonly description: "Further level of headings for complex document";
        } | {
            readonly name: "heading5";
            readonly id: "heading5";
            readonly type: "text";
            readonly htmlTagName: "h5";
            readonly className: "-heading5";
            readonly disguise: string;
            readonly isImplicit: true;
            readonly isCustom: false;
            readonly title: "Heading 5";
            readonly label: "Heading 5";
            readonly description: "The deepest heading level available";
        } | {
            readonly name: "listItem";
            readonly id: "listItem";
            readonly type: "text";
            readonly htmlTagName: "li";
            readonly label: "List item";
            readonly disguise: "paragraph";
            readonly className: "-list-item";
            readonly isImplicit: true;
            readonly isCustom: false;
        } | {
            readonly name: "link";
            readonly id: "link";
            readonly type: "inline";
            readonly label: "Link";
            readonly htmlTagName: keyof HTMLElementTagNameMap;
            readonly className: "-link";
            readonly disguise: string;
            readonly placeholder: "New link";
            readonly isImplicit: true;
            readonly canBeLink: true;
            readonly isCustom: false;
            readonly attributes: {
                readonly linkHref: "";
            };
            readonly title: "Links";
            readonly description: "Links inside text elements";
        } | {
            readonly name: "button";
            readonly id: "button";
            readonly type: "inline";
            readonly label: "Button";
            readonly htmlTagName: "button";
            readonly className: "-button";
            readonly disguise: string;
            readonly placeholder: "Click me";
            readonly isImplicit: true;
            readonly isCustom: false;
            readonly canBeLink: true;
            readonly title: "Buttons";
            readonly description: "Clickable element used for user interactions";
        } | {
            readonly name: "badge";
            readonly id: "badge";
            readonly type: "inline";
            readonly label: "Badge";
            readonly className: "-inline--badge";
            readonly disguise: string;
            readonly placeholder: "Badge text";
            readonly isImplicit: false;
            readonly htmlTagName: keyof HTMLElementTagNameMap;
            readonly isCustom: true;
            readonly canBeLink: true;
            readonly title: "Badges";
            readonly description: "Element represeting status and other meta information";
        } | {
            readonly name: "time";
            readonly id: "time";
            readonly type: "inline";
            readonly label: "Date";
            readonly htmlTagName: "time";
            readonly className: "-inline--time";
            readonly disguise: string;
            readonly placeholder: "2023-01-01";
            readonly isImplicit: true;
            readonly isCustom: true;
            readonly canBeLink: true;
            readonly description: "Present mapped date in human readable format";
        } | {
            readonly name: "inline";
            readonly id: "inline";
            readonly type: "inline";
            readonly label: "Generic inline element";
            readonly className: "-inline";
            readonly disguise: string;
            readonly isImplicit: false;
            readonly placeholder: "Inline element";
            readonly canBeLink: true;
            readonly htmlTagName: keyof HTMLElementTagNameMap;
            readonly isCustom: false;
        } | {
            readonly type: "inline";
            readonly id: "var";
            readonly label: "Variable";
            readonly name: "var";
            readonly className: string;
            readonly placeholder: "Double click to edit";
            readonly disguise: string;
            readonly htmlTagName: keyof HTMLElementTagNameMap;
            readonly isImplicit: true;
            readonly isCustom: false;
            readonly description: "Inline placement to be substituted with mapped text";
            readonly attributes: {
                readonly 'data-path': "";
            };
        } | {
            readonly name: "section";
            readonly id: "section";
            readonly type: "block";
            readonly label: "Section";
            readonly htmlTagName: "section";
            readonly className: "-section";
            readonly disguise: string;
            readonly isImplicit: true;
            readonly isCustom: false;
            readonly title: "Sections";
            readonly description: "Themeable essential boxes represeting top-level chunks of a component";
            readonly priority: -1;
            readonly attributes: {
                readonly class: "-grid--custom";
            };
        } | {
            readonly name: "card";
            readonly id: "card";
            readonly type: "block";
            readonly label: "Card";
            readonly htmlTagName: keyof HTMLElementTagNameMap;
            readonly className: "-card";
            readonly disguise: string;
            readonly isImplicit: true;
            readonly canBeLink: true;
            readonly isCustom: false;
            readonly title: "Cards";
            readonly description: "Themable nestable boxes that provide visual framing for content";
            readonly priority: -1;
        } | {
            readonly name: "blockquote";
            readonly id: "blockquote";
            readonly type: "block";
            readonly label: "Block quote";
            readonly htmlTagName: "blockquote";
            readonly className: "-blockquote";
            readonly disguise: string;
            readonly isImplicit: true;
            readonly canBeLink: true;
            readonly isCustom: false;
            readonly title: "Block quotes";
            readonly description: "Quotes, tweets and rich styled text references";
            readonly priority: -1;
        } | {
            readonly name: "form";
            readonly id: "form";
            readonly type: "block";
            readonly label: "Form";
            readonly group: "external";
            readonly elementName: "component";
            readonly className: "-component--form";
            readonly disguise: "media";
            readonly canBeLink: false;
            readonly htmlTagName: keyof HTMLElementTagNameMap;
            readonly isImplicit: true;
            readonly isCustom: true; /** Return a current draft or create a new draft bumping up the revision */
            readonly description: "Form created in Sitecore Forms";
            readonly priority: 2;
        } | {
            readonly name: "component";
            readonly id: "component";
            readonly type: "block";
            readonly label: "Component";
            readonly group: "external";
            readonly className: "-component";
            readonly disguise: "media";
            readonly canBeLink: false;
            readonly htmlTagName: keyof HTMLElementTagNameMap;
            readonly isImplicit: true;
            readonly isCustom: false;
            readonly description: "Nested web, react or FEAAS component";
            readonly priority: 1;
        } | {
            readonly name: "embed";
            readonly id: "embed";
            readonly type: "block";
            readonly group: "block";
            readonly label: "HTML Block";
            readonly className: "-embed";
            readonly disguise: "media";
            readonly canBeLink: true;
            readonly htmlTagName: keyof HTMLElementTagNameMap;
            /** Register a new revision of a stylesheet with given properties based on this stylesheet  */
            readonly isImplicit: true;
            readonly isCustom: false;
            readonly description: "Raw HTML content embedded into the component";
            readonly priority: 3;
        } | {
            readonly name: "media";
            readonly id: "media";
            readonly type: "block";
            readonly label: "Image";
            readonly htmlTagName: keyof HTMLElementTagNameMap;
            readonly isImplicit: true;
            readonly isCustom: false;
            readonly canBeLink: true;
            readonly className: "-block--media";
            readonly disguise: string;
            readonly title: "Image blocks";
            readonly description: "Image with optional accompanying text content";
            readonly priority: -1;
            readonly attributes: {
                readonly image: "";
            };
        } | {
            readonly name: "block";
            readonly id: "block";
            readonly type: "block";
            readonly label: "Generic block element";
            readonly className: "-block";
            readonly disguise: string;
            readonly isImplicit: false;
            readonly canBeLink: true;
            readonly htmlTagName: keyof HTMLElementTagNameMap;
            readonly isCustom: false;
            readonly priority: 1;
        } | {
            readonly name: "spacer";
            readonly id: "spacer";
            readonly type: "block";
            readonly label: "Spacer";
            readonly className: "-spacer";
            readonly disguise: string;
            readonly canBeLink: false;
            readonly htmlTagName: keyof HTMLElementTagNameMap;
            readonly isImplicit: true;
            readonly isCustom: false;
            readonly description: "Flexible spacer that expands to take all available space";
            readonly priority: 2;
        } | {
            readonly name: "container";
            readonly id: "container";
            readonly type: "block";
            readonly label: "Block";
            readonly className: "-container";
            readonly disguise: string;
            readonly canBeLink: true;
            readonly isImplicit: true;
            readonly isCustom: false;
            readonly htmlTagName: keyof HTMLElementTagNameMap;
            readonly priority: -1;
            readonly description: "Unstyled container for layout purposes";
        };
    };
    get Set(): typeof Style.Set;
}
export { StylesheetModel as Sheet };
//# sourceMappingURL=stylesheets.d.ts.map