import SDKAdapter from '../core/adapter.js';
import { SDKCollection } from '../core/collection.js';
import { SDKModel } from '../core/model.js';
import { Atleast, Optional, RequiredKeys } from '../utils/typescript.js';
import { SDKModelSchema } from '../core/schema.js';
import type { SDK } from '../sdk.js';
import * as Style from '../style/index.js';
import type { CollectionModel } from './collections.js';
import type { ComponentModel } from './components.js';
import type { LibraryModel, NanoId } from './libraries.js';
import { DataOptionsById, DatasourceModel } from './datasources.js';
import type { VersionBundle } from './version-bundles.js';
/** Defines the schema for Version. */
export interface Version {
    /** The identifier of the Library that Version belongs to */
    libraryId: NanoId;
    /** The reference identifer of the Version */
    id: NanoId;
    /** The identifier of the Component that Version belongs to */
    componentId: NanoId;
    /**
     * The name of the Version.
     *
     * @minLength 1
     * @maxLength 128
     */
    name: string;
    /** Textual description of the Version. */
    description?: string;
    /** The status of the Version */
    status: 'published' | 'draft' | 'staged' | 'saved';
    /**
     * The order index of the Version.
     *
     * @minLength 1
     * @maxLength 64
     */
    orderIdx?: string;
    /**
     * The numerc revision of the Version.
     *
     * @minimum 0
     * @TJS-type integer
     */
    revision: number;
    /** The model presentation of the Version */
    model?: string;
    /**
     * The view presentation of the Version.
     *
     * @minLength 1
     */
    view: string;
    /** The difference between current and previous version of the Version */
    diff?: object;
    /**
     * List of ids of applicable breakpoints. Null if all breakpoints are applicable.
     *
     * @nullable
     */
    breakpointIds?: string[];
    /**
     * The identifiers of the datasources for this Version.
     *
     * @items.type string
     */
    datasourceIds?: string[];
    /** Allow version to be embeddable on its own, not as a part of responsive bundle */
    isEmbeddable?: boolean;
    /**
     * Id of a version that was used as a base for the fork.
     *
     * @nullable
     */
    forkOriginId?: NanoId;
    /**
     * Revision of a version that was used as a base for the fork.
     *
     * @nullable
     * @minimum 0
     * @TJS-type integer
     */
    forkOriginRevision?: number;
    /**
     * Time when the version was first forked.
     *
     * @nullable
     */
    forkedAt?: Date;
    /**
     * Array of extra css classes applied to the version.
     *
     * @nullable
     */
    classList?: string[];
    /** The timestamp when the Version was created at */
    createdAt?: Date;
    /**
     * The timestamp when the Version was deleted at.
     *
     * @nullable
     */
    deletedAt?: Date;
    /** The timestamp when the Version was last modified at */
    modifiedAt?: Date;
    /** The User who last modified the Version */
    modifiedBy?: object;
}
export interface VersionSnapshot {
    id: VersionModel['id'];
    head?: VersionModel;
    previous?: VersionModel;
    draft?: VersionModel;
    saved?: VersionModel;
    staged?: VersionModel;
    published?: VersionModel;
    revisions: VersionModel[];
}
export declare class Versions extends SDKAdapter<Version> {
    search(version: Pick<VersionParams, 'libraryId'>, query?: any): Promise<Version[]>;
    get(version: Pick<VersionParams, 'libraryId' | 'collectionId' | 'componentId' | 'id' | 'status' | 'revision'>): Promise<Version>;
    fetch(version: Pick<VersionParams, 'libraryId' | 'collectionId' | 'componentId'>): Promise<Version[]>;
    post(version: Optional<Version, 'componentId' | 'id' | 'revision'> & Atleast<VersionParams, 'libraryId' | 'collectionId' | 'componentId'>): Promise<any>;
    put(version: Optional<Version, 'componentId' | 'revision'> & Atleast<VersionParams, 'libraryId' | 'collectionId' | 'componentId'>): Promise<any>;
    delete(version: Atleast<VersionParams, 'libraryId' | 'collectionId' | 'componentId'>): Promise<any>;
    prune({ libraryId, componentId, id }?: Partial<Pick<Version, 'libraryId' | 'componentId' | 'id'>>): void;
}
declare function getVersionDefaults(sdk: SDK): {
    id: string;
    orderIdx: string;
    revision: number;
    createdAt: Date;
    modifiedAt: Date;
    deletedAt: Date;
    modifiedBy: {
        id?: undefined;
        name?: undefined;
        givenName?: undefined;
        familyName?: undefined;
        email?: undefined;
        picture?: undefined;
        nickname?: undefined;
    } | {
        id: string;
        name: string;
        givenName: string;
        familyName: string;
        email: string;
        picture: string;
        nickname: string;
    };
    datasourceIds: string[];
    breakpointIds: string[];
    classList: string[];
    view: string;
    model: string;
    diff: {};
    status: string;
};
export type VersionMinimal<T extends keyof Version = never> = RequiredKeys<Version, typeof getVersionDefaults, T>;
export interface VersionImplicit {
    library?: LibraryModel;
    libraryId: LibraryModel['id'];
    collectionId?: CollectionModel['id'];
    collection?: CollectionModel;
    componentId: ComponentModel['id'];
    component?: ComponentModel;
    bundle?: VersionBundle;
    bundleId?: VersionBundle['id'];
}
export interface VersionParams extends Version, VersionImplicit {
}
export interface VersionModel extends VersionParams {
}
export declare class VersionModel extends SDKModel<Version, VersionMinimal, VersionImplicit> implements VersionParams {
    error?: Error;
    get adapter(): Versions;
    getId(): string;
    getHiddenProps(): string[];
    getDefaults(sdk: SDK): {
        id: string;
        orderIdx: string;
        revision: number;
        createdAt: Date;
        modifiedAt: Date;
        deletedAt: Date;
        modifiedBy: {
            id?: undefined;
            name?: undefined;
            givenName?: undefined;
            familyName?: undefined;
            email?: undefined;
            picture?: undefined;
            nickname?: undefined;
        } | {
            id: string;
            name: string;
            givenName: string;
            familyName: string;
            email: string;
            picture: string;
            nickname: string;
        };
        datasourceIds: string[];
        breakpointIds: string[];
        classList: string[];
        view: string;
        model: string;
        diff: {};
        status: string;
    };
    getThumbnailURL(): string;
    defineProperties(): void;
    /** View needs to be wrapped into <section> */
    _view: string;
    get view(): string;
    isViewEmpty(): RegExpMatchArray;
    /** Normalizes html, extracts used datasources */
    set view(value: string);
    /** Extract all datasource ids from view used in data-path-* attributes */
    getDatasourcesFromView(view: string): string[];
    /**
     * Transform view minimally to help reduce amount of revisions. It tries to bring HTML to a canonical version, so that
     * things like different order of classes will not be considered a different version
     */
    normalizeView(view: string): string;
    /** Combine view, classes and themes */
    getHTML(extraClasses?: string[]): string;
    /** Get latest revisions of a version matching given statuses (or first) */
    getCurrent(statuses?: Version['status'][] | Version['status']): VersionModel;
    /** Get settings for all datasoruces used in version */
    getDataSettings(): DataOptionsById | null;
    /** Get current draft or create a new one */
    getDraft(): VersionModel;
    /**
     * Use current revision as a basis to create an entirely new one-off version. Used for for in-place editing of
     * components.
     */
    fork(id?: string): VersionModel;
    /** Return to the shared version of a component. The fork will be marked as deleted. */
    unfork(): VersionModel;
    /** Create new version based on the current with new ID */
    produceNewVersion(data?: Partial<Version>): VersionModel;
    /**
     * Creates new instance of a version revision if any data has changed.
     *
     * ! Replaces it in the collection to notify observers
     */
    changeAndNotify(data: Partial<Version>): VersionModel;
    /** Move version after another matching given id (or to the top if null) */
    move(anchorId: Version['id'], position?: 'before' | 'after'): VersionModel;
    /** Soft-delete saved version, discards draft */
    archive(): VersionModel;
    compareProp(property: keyof this, left: any, right: any): boolean;
    /** Soft-undelete version */
    restore(): VersionModel;
    /** Set version size breakpoints and commit as new revision */
    setBreakpointIds(breakpointIds: Version['breakpointIds']): void;
    /** Creates new version with provided combo */
    setClassList(classList: string[]): VersionModel;
    /** Set current view/model data to version. Ensures that there is a draft, in case given data has changes */
    commitData({ view, model }: {
        view: Version['view'];
        model: Version['model'];
    }): VersionModel;
    /** Commits given data a new saved revision if there were an changes */
    saveData(data?: Atleast<Version, 'view' | 'model' | 'breakpointIds'>): this;
    /** Create a separate version starting from first revision, but based on given one */
    duplicate(): VersionModel;
    /** Change name of version (does not create new revision) */
    rename(name: Version['name']): void;
    setIsEmbeddable(isEmbeddable: Version['isEmbeddable']): void;
    /** Create a staged revision based on current one. Ensures that it's not soft-deleted */
    stage(): VersionModel;
    /** Soft-deletes staged revision */
    unstage(): VersionModel;
    /** Create a published revision based on current one */
    /** Ensures that it's not soft-deleted */
    publish(): VersionModel;
    /** Soft-deletes published revision */
    unpublish(): VersionModel;
    /** Discards the draft, removes it from revision list */
    discard(): this;
    /** Revert current changes to last published/staged one */
    revertTo(revision: VersionModel): VersionModel;
    toProperties(): {
        view: string;
        model: string;
        breakpointIds: string[];
    };
    revert(): VersionModel;
    isDifferentFrom(other: Version): boolean;
    getBreakpoints(): Style.RuleWithProps<"breakpoint", {
        minWidth: number;
        maxWidth: number;
        query: string;
    }>[];
    getDatasources(): DatasourceModel[];
    matchBreakpoint(breakpoint: Style.Rule<'breakpoint'>, targetWidth: number): boolean;
    matchesSearch(search: string): boolean;
    static getOrderedBreakpoints(targetWidth: number, breakpointsArray: Style.RuleWithProps<'breakpoint', {
        minWidth: number;
        maxWidth: number;
        query: string;
    }>[]): Style.RuleWithProps<"breakpoint", {
        minWidth: number;
        maxWidth: number;
        query: string;
    }>[];
    getMatchingBreakpoint(targetWidth: number): Style.RuleWithProps<"breakpoint", {
        minWidth: number;
        maxWidth: number;
        query: string;
    }>;
    getUsedComboIds(): string[];
    getLastPushed(): VersionModel;
    needsStaging(): boolean;
    needsPublishing(): boolean;
    isDeleted(): boolean;
    getThemeSlug(): string;
    toStyleContext(): {
        name: string;
        getAttribute: (name: string) => string;
    };
    static getOrderedVersions(versions: VersionModel[]): VersionModel[];
    static getVersionOrderIdx(left: Version, right: Version): string;
    sortCompare(other: this): number;
    /** Find all unique in html and generate new ones, fix references */
    static getUniqueHTML(view: string, map?: Record<string, string>): string;
    static statuses: string[];
    static sortVersionRevisions(revisions: VersionModel[]): VersionModel[];
    static reduceVersionRevisions(revisions: VersionModel[]): VersionSnapshot[];
    static onCollectionChange(versions: SDKCollection<typeof VersionModel>): VersionModel[];
    getPath(): string;
    getSourcePath(): string;
    static get schema(): SDKModelSchema<typeof VersionModel, Version>;
}
export {};
//# sourceMappingURL=versions.d.ts.map