import type * as Style from '../index.js';
/**
 * # Class list methods
 *
 * Application and resolution of style rules is done in context of class list. It is enough to tell which element, rule
 * and combo styles are applied to an element. In cases where combo styles need to be resolved it may require theme
 * context class list as well. For elements that do not have actual class attribute assigned like `<p>` or `<section>`
 * the class list is automatically augmented with _implicit class_ by @see Context#getClassName
 */
/**
 * Parses a class name string into a normalized class list.
 *
 * @param {string} string  - A string containing class names separated by spaces.
 * @returns {string[]} An array of normalized class names.
 * @example
 *
 *   fromClassName('classA  classB') // ['classA', 'classB']
 *   fromClassName('classA classB classA') // ['classA', 'classB']
 *
 */
export declare function fromClassName(string: string): string[];
/**
 * Adds a class or a list of classes to the class list while keeping it normalized.
 *
 * @param {string[]}          classList  - The existing list of class names.
 * @param {string | string[]} newClass   - The new class name(s) to add.
 * @returns {string[]} An array of class names with the new class(es) added.
 * @example
 *
 *   concat(['classA', 'classB'], 'classC') // ['classA', 'classB', 'classC']
 *   concat(['classA', 'classB'], ['classC', 'classD']) // ['classA', 'classB', 'classC', 'classD']
 *   concat(['classA', 'classB'], ['classC', 'classB']) // ['classA', 'classB', 'classC']
 *
 */
export declare function concat(classList: string[], newClass: string | string[]): string[];
/**
 * Normalizes a class list by sorting it and removing duplicates.
 *
 * @param {string[]} classList  - The existing list of class names.
 * @returns {string[]} An array of class names sorted and without duplicates.
 * @example
 *
 *   normalize(['classA', 'classB', 'classA']) // ['classA', 'classB']
 *   normalize(['classB', 'classA', 'classC', 'classB']) // ['classA', 'classB', 'classC']
 *
 */
export declare function normalize(classList: string[]): string[];
/**
 * Sorts a class list in alphabetical order, with the element class name at the beginning.
 *
 * @param {string[]} classList  - The existing list of class names.
 * @returns {string[]} An array of class names sorted as described.
 * @example
 *
 *   sort(['classB', '-elementA', 'classA']) // ['-elementA', 'classA', 'classB']
 *   sort(['classC', 'classA', '-elementB', 'classB']) // ['-elementB', 'classA', 'classB', 'classC']
 *
 */
export declare function sort(classList: string[]): string[];
/**
 * Swaps an element style in the class list. When changing element styles, also resets aspect styles. If passed an empty
 * style, resets the class list to the generic class name, effectively making it subject to theme style.
 *
 * @param {string[]}                 classList  - The existing list of class names.
 * @param {Style.Rule | Style.Empty} style      - The style to set.
 * @returns {string[]} An array of class names with the style set.
 * @example
 *
 * setStyle(['-block', 'classA'], styleObject) // ['-block--newStyle', 'classA']
 *   setStyle(['-block--styleA', 'classA'], {type: 'block}) // ['-elementA', 'classA']
 *
 */
export declare function setStyle(classList: string[], style: Style.Rule | Style.Empty): string[];
/**
 * Swaps a theme in the class list and removes the previously assigned combos.
 *
 * @param {string[]}            classList  - The existing list of class names.
 * @param {Style.Rule<'theme'>} theme      - The theme to set.
 * @returns {string[]} An array of class names with the theme set.
 * @example
 *
 *   setTheme(['-theme--oldTheme', 'classA'], themeObject) // ['-theme--newTheme', 'classA']
 *   setTheme(['-theme--oldTheme', '--use--combo'], themeObject) // ['-theme--newTheme']
 *   setTheme(['classA', 'classB'], themeObject) // ['-theme--newTheme', 'classA', 'classB']
 *
 */
export declare function setTheme(classList: string[], theme: Style.Rule<'theme'>): string[];
/**
 * Applies a -use--combo.id to the class list and removes deleted combos and combos that belong to the same category as
 * the selected combo. It will clean up theme references to theme combos that don't exist as well.
 *
 * @param {string[]}          classList  - The existing list of class names.
 * @param {Style.Theme.Combo} combo      - The combo to set.
 * @param {Style.Rule[]}      styles     - The array of style rules.
 * @returns {string[]} An array of class names with the combo set.
 * @example
 *
 *   setThemeCombo(['-use--oldParagraphCombo', 'classA'], comboObject, stylesArray) // ['-use--newParagraphCombo', 'classA']
 *   setThemeCombo(['classA', 'classB'], comboObject, stylesArray) // ['-use--newCombo', 'classA', 'classB']
 *
 */
export declare function setThemeCombo(classList: string[], combo: Style.Theme.Combo, styles: Style.Rule[]): string[];
/**
 * Removes a specific theme combo from the class list by its id.
 *
 * @param {string[]}          classList  - The existing list of class names.
 * @param {Style.Theme.Combo} combo      - The combo to be removed.
 * @returns {string[]} An array of class names with the specified combo removed.
 * @example
 *
 *   removeThemeCombo(['-use--oldParagraphCombo', 'classA'], oldParagraphCombo) // ['classA']
 *   removeThemeCombo(['classA', 'classB'], nonExistingCombo) // ['classA', 'classB']
 *
 */
export declare function removeThemeCombo(classList: string[], combo: Style.Theme.Combo): string[];
/**
 * Swaps, applies, or removes a combo from the class list. Combos are identified by the `--uid` format of their class
 * names.
 *
 * @param {string[]}          classList  - The existing list of class names.
 * @param {Style.Theme.Combo} combo      - The combo to set.
 * @returns {string[]} An array of class names with the combo set.
 * @example
 *
 *   setCombo(['--oldCombo', 'classA'], comboObject) // ['--newCombo', 'classA']
 *   setCombo(['classA', 'classB'], comboObject) // ['--newCombo', 'classA', 'classB']
 *
 */
export declare function setCombo(classList: string[], combo: Style.Theme.Combo): string[];
/**
 * Given a set of combos, returns the applied combo or the default one.
 *
 * @param {string[]}            classList  - The existing list of class names.
 * @param {Style.Theme.Combo[]} combos     - The array of combos.
 * @returns {Style.Theme.Combo} The applied or default combo.
 * @example
 *
 *   findCombo(['--comboA', 'classA'], combosArray) // comboA_Object
 *   findCombo(['classA', 'classB'], combosArray) // defaultCombo_Object
 *
 */
export declare function findCombo(classList: string[], combos: Style.Theme.Combo[]): {
    id: string;
    refId: string;
    fillId: string;
    decorationId: string;
    spacingId: string;
    typographyId: string;
    paletteId: string;
    isDefault: boolean;
};
/**
 * Finds the element class name in a list and simplifies it to remove the specific style.
 *
 * @param {string[]} classList  - The existing list of class names.
 * @returns {string} The generic class name of the element.
 * @example
 *
 *   getElementGenericClassName(['-paragraph--styleA', 'classA']) // '-paragraph'
 *   getElementGenericClassName(['-block--hero--styleB', 'classB', 'classC']) // '-block--hero'
 *
 */
export declare function getElementGenericClassName(classList: string[]): string;
/**
 * Finds the element class name in a list. If there're multiple class that can be definitions, prioritize them.
 *
 * @param {string[]} classList  - The existing list of class names.
 * @returns {string} The class name of the element.
 * @example
 *
 *   getElementClassName(['-card', 'classA']) // '-element
 *   getElementClassName(['-card--styleA', 'classA']) // '-elementA--styleA'
 *   getElementClassName(['-inline--badge--styleB', 'classB', 'classC']) // '-inline--badge--styleB'
 *
 */
export declare function getElementClassName(classList: string[], definitions?: Style.ElementDefinition[]): string;
/**
 * Finds the definition for a class list.
 *
 * @param {string[]} classList  - The list of class names.
 * @returns {Style.ElementDefinition} The corresponding element definition.
 * @example
 *
 *   getDefinition(['-card--styleA', 'classA']) // Style.Rule.Element
 *   getDefinition(['-inline--badge', 'classB']) // Style.Rule.Element
 *
 */
export declare function getDefinition(classList: string[]): {
    readonly name: "paragraph";
    readonly id: "paragraph";
    readonly type: "text";
    readonly htmlTagName: "p";
    readonly className: "-paragraph";
    readonly disguise: string;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly placeholder: "Text element";
    readonly title: "Paragraphs";
    readonly label: "Paragraph";
    readonly description: "Regular text element used for all text and copy that isn't heading";
} | {
    readonly name: "heading1";
    readonly id: "heading1";
    readonly type: "text";
    readonly htmlTagName: "h1";
    readonly className: "-heading1";
    readonly disguise: string;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly title: "Heading 1";
    readonly label: "Heading 1";
    readonly description: "Main heading of the document, used for largest of headings";
} | {
    readonly name: "heading2";
    readonly id: "heading2";
    readonly type: "text";
    readonly htmlTagName: "h2";
    readonly className: "-heading2";
    readonly disguise: string;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly title: "Heading 2";
    readonly label: "Heading 2";
    readonly description: "Second level of headings";
} | {
    readonly name: "heading3";
    readonly id: "heading3";
    readonly type: "text";
    readonly htmlTagName: "h3";
    readonly className: "-heading3";
    readonly disguise: string;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly title: "Heading 3";
    readonly label: "Heading 3";
    readonly description: "Third level of headings often used for taglines";
} | {
    readonly name: "heading4";
    readonly id: "heading4";
    readonly type: "text";
    readonly htmlTagName: "h4";
    readonly className: "-heading4";
    readonly disguise: string;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly title: "Heading 4";
    readonly label: "Heading 4";
    readonly description: "Further level of headings for complex document";
} | {
    readonly name: "heading5";
    readonly id: "heading5";
    readonly type: "text";
    readonly htmlTagName: "h5";
    readonly className: "-heading5";
    readonly disguise: string;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly title: "Heading 5";
    readonly label: "Heading 5";
    readonly description: "The deepest heading level available";
} | {
    readonly name: "listItem";
    readonly id: "listItem";
    readonly type: "text";
    readonly htmlTagName: "li";
    readonly label: "List item";
    readonly disguise: "paragraph";
    readonly className: "-list-item";
    readonly isImplicit: true;
    readonly isCustom: false;
} | {
    readonly name: "link";
    readonly id: "link";
    readonly type: "inline";
    readonly label: "Link";
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly className: "-link";
    readonly disguise: string;
    readonly placeholder: "New link";
    readonly isImplicit: true;
    readonly canBeLink: true;
    readonly isCustom: false;
    readonly attributes: {
        readonly linkHref: "";
    };
    readonly title: "Links";
    readonly description: "Links inside text elements";
} | {
    readonly name: "button";
    readonly id: "button";
    readonly type: "inline";
    readonly label: "Button";
    readonly htmlTagName: "button";
    readonly className: "-button";
    readonly disguise: string;
    readonly placeholder: "Click me";
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly canBeLink: true;
    readonly title: "Buttons";
    readonly description: "Clickable element used for user interactions";
} | {
    readonly name: "badge";
    readonly id: "badge";
    readonly type: "inline";
    readonly label: "Badge";
    readonly className: "-inline--badge";
    readonly disguise: string;
    readonly placeholder: "Badge text";
    readonly isImplicit: false;
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly isCustom: true;
    readonly canBeLink: true;
    readonly title: "Badges";
    readonly description: "Element represeting status and other meta information";
} | {
    readonly name: "time";
    readonly id: "time";
    readonly type: "inline";
    readonly label: "Date";
    readonly htmlTagName: "time";
    readonly className: "-inline--time";
    readonly disguise: string;
    readonly placeholder: "2023-01-01";
    readonly isImplicit: true;
    readonly isCustom: true;
    readonly canBeLink: true;
    readonly description: "Present mapped date in human readable format";
} | {
    readonly name: "inline";
    readonly id: "inline";
    readonly type: "inline";
    readonly label: "Generic inline element";
    readonly className: "-inline";
    readonly disguise: string;
    readonly isImplicit: false;
    readonly placeholder: "Inline element";
    readonly canBeLink: true;
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly isCustom: false;
} | {
    readonly type: "inline";
    readonly id: "var";
    readonly label: "Variable";
    readonly name: "var";
    readonly className: string;
    readonly placeholder: "Double click to edit";
    readonly disguise: string;
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly description: "Inline placement to be substituted with mapped text";
    readonly attributes: {
        readonly 'data-path': "";
    };
} | {
    readonly name: "section";
    readonly id: "section";
    readonly type: "block";
    readonly label: "Section";
    readonly htmlTagName: "section";
    readonly className: "-section";
    readonly disguise: string; /**
     * Swaps a theme in the class list and removes the previously assigned combos.
     *
     * @param {string[]}            classList  - The existing list of class names.
     * @param {Style.Rule<'theme'>} theme      - The theme to set.
     * @returns {string[]} An array of class names with the theme set.
     * @example
     *
     *   setTheme(['-theme--oldTheme', 'classA'], themeObject) // ['-theme--newTheme', 'classA']
     *   setTheme(['-theme--oldTheme', '--use--combo'], themeObject) // ['-theme--newTheme']
     *   setTheme(['classA', 'classB'], themeObject) // ['-theme--newTheme', 'classA', 'classB']
     *
     */
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly title: "Sections";
    readonly description: "Themeable essential boxes represeting top-level chunks of a component";
    readonly priority: -1;
    readonly attributes: {
        readonly class: "-grid--custom";
    };
} | {
    readonly name: "card";
    readonly id: "card";
    readonly type: "block";
    readonly label: "Card";
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly className: "-card";
    readonly disguise: string;
    readonly isImplicit: true;
    readonly canBeLink: true;
    readonly isCustom: false;
    readonly title: "Cards";
    readonly description: "Themable nestable boxes that provide visual framing for content";
    readonly priority: -1;
} | {
    readonly name: "blockquote";
    readonly id: "blockquote";
    readonly type: "block";
    readonly label: "Block quote";
    readonly htmlTagName: "blockquote";
    readonly className: "-blockquote";
    readonly disguise: string;
    readonly isImplicit: true;
    readonly canBeLink: true;
    readonly isCustom: false;
    /**
     * Applies a -use--combo.id to the class list and removes deleted combos and combos that belong to the same category as
     * the selected combo. It will clean up theme references to theme combos that don't exist as well.
     *
     * @param {string[]}          classList  - The existing list of class names.
     * @param {Style.Theme.Combo} combo      - The combo to set.
     * @param {Style.Rule[]}      styles     - The array of style rules.
     * @returns {string[]} An array of class names with the combo set.
     * @example
     *
     *   setThemeCombo(['-use--oldParagraphCombo', 'classA'], comboObject, stylesArray) // ['-use--newParagraphCombo', 'classA']
     *   setThemeCombo(['classA', 'classB'], comboObject, stylesArray) // ['-use--newCombo', 'classA', 'classB']
     *
     */
    readonly title: "Block quotes";
    readonly description: "Quotes, tweets and rich styled text references";
    readonly priority: -1;
} | {
    readonly name: "form";
    readonly id: "form";
    readonly type: "block";
    readonly label: "Form";
    readonly group: "external";
    readonly elementName: "component";
    readonly className: "-component--form";
    readonly disguise: "media";
    readonly canBeLink: false;
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly isImplicit: true;
    readonly isCustom: true;
    readonly description: "Form created in Sitecore Forms";
    readonly priority: 2;
} | {
    readonly name: "component";
    readonly id: "component";
    readonly type: "block";
    readonly label: "Component";
    readonly group: "external";
    readonly className: "-component";
    readonly disguise: "media";
    readonly canBeLink: false;
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly description: "Nested web, react or FEAAS component";
    readonly priority: 1;
} | {
    readonly name: "embed";
    readonly id: "embed";
    readonly type: "block";
    readonly group: "block";
    readonly label: "HTML Block";
    readonly className: "-embed";
    readonly disguise: "media";
    readonly canBeLink: true;
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly description: "Raw HTML content embedded into the component";
    readonly priority: 3;
} | {
    readonly name: "media";
    readonly id: "media";
    readonly type: "block";
    readonly label: "Image";
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly canBeLink: true;
    readonly className: "-block--media";
    readonly disguise: string;
    readonly title: "Image blocks";
    readonly description: "Image with optional accompanying text content";
    readonly priority: -1;
    readonly attributes: {
        readonly image: "";
    };
} | {
    readonly name: "block";
    readonly id: "block";
    readonly type: "block";
    readonly label: "Generic block element";
    readonly className: "-block";
    readonly disguise: string;
    readonly isImplicit: false;
    readonly canBeLink: true;
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly isCustom: false;
    readonly priority: 1;
} | {
    readonly name: "spacer";
    readonly id: "spacer";
    readonly type: "block";
    readonly label: "Spacer";
    readonly className: "-spacer";
    readonly disguise: string;
    readonly canBeLink: false;
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly description: "Flexible spacer that expands to take all available space";
    readonly priority: 2;
} | {
    readonly name: "container";
    readonly id: "container";
    readonly type: "block";
    readonly label: "Block";
    readonly className: "-container";
    readonly disguise: string;
    readonly canBeLink: true;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly priority: -1;
    readonly description: "Unstyled container for layout purposes";
};
/**
 * Checks if a class list contains an Element style class name.
 *
 * @param {string[]}           classList  - The list of class names.
 * @param {Style.Rule.Element} style      - The Element style rule.
 * @returns {boolean} True if the Element style class name is in the class list, false otherwise.
 * @example
 *
 *   matchesElement(['-card--styleA', 'classA'], elementAStyle) // true
 *   matchesElement(['-card--styleB', 'classB'], elementAStyle) // false
 *
 */
export declare function matchesElement(classList: string[], style: Style.Rule.Element): boolean;
/**
 * Checks if a class list contains class names that match a given definition.
 *
 * @param {string[]}                classList   - The list of class names.
 * @param {Style.ElementDefinition} definition  - The element definition to match.
 * @returns {boolean} True if a class name in the list matches the definition, false otherwise.
 * @example
 *
 *   matchesDefinition(['-card--styleA', 'classA'], cardDefinition) // true
 *   matchesDefinition(['-section--styleB', 'classB'], cardDefinition) // false
 *
 */
export declare function matchesDefinition(classList: string[], definition: Style.ElementDefinition): boolean;
/**
 * Checks if a combo id in the class list matches an element.
 *
 * @param {string[]}     classList  - The list of class names.
 * @param {string}       comboId    - The combo id to match.
 * @param {Style.Rule[]} styles     - The set of style rules.
 * @returns {boolean} True if the combo id matches the element, false otherwise.
 * @example
 *
 *   matchesComboId(['-card--styleA'], cardCombo.id, styles) // true
 *   matchesComboId(['-section--styleB'], cardCombo.id, styles) // false
 *
 */
export declare function matchesComboId(classList: string[], comboId: string, styles: Style.Rule[]): boolean;
/**
 * Finds the theme slug in a class list.
 *
 * @param {string[]} classList  - The list of class names.
 * @returns {string} The theme slug if found, undefined otherwise.
 * @example
 *
 *   getThemeSlug(['-theme--default', '-use--deadbeef']) // 'default'
 *   getThemeSlug(['-theme--custom', '-use--deadbeef']) // 'custom'
 *   getThemeSlug(['-use--deadbeef']) // undefined
 *
 */
export declare function getThemeSlug(classList: string[]): string;
/**
 * Lists all combo ids in a class list of theme context.
 *
 * @param {string[]} classList  - The list of class names.
 * @returns {string[]} An array of combo ids found in the class list.
 * @example
 *
 *   getThemeComboIds(['-theme--default', '-use--deadbeef']) // ['deadbeef']
 *   getThemeComboIds(['-theme--default', '-use--deadbeef', '-use--baddad']) // ['deadbeef', 'baddad']
 *   getThemeComboIds(['-theme--default']) // []
 *
 */
export declare function getThemeComboIds(classList: string[]): string[];
/**
 * Lists all combo ids in a class list of element context.
 *
 * @param {string[]} classList  - The list of class names.
 * @returns {string | null} The combo id if found, null otherwise.
 * @example
 *
 *   getComboId(['-section', '--deadbeef']) // 'deadbeef'
 *   getComboId(['-section']) // null
 *
 */
export declare function getComboId(classList: string[]): string;
/**
 * Enriches a class list with an implicit class based on the context name.
 *
 * @param {string[]} classList    - The list of class names.
 * @param {string}   contextName  - The context name.
 * @returns {string[]} The expanded class list with the implicit class added if applicable.
 * @example
 *
 *   expand(['abc'], 'section') // ['-section', 'abc']
 *   expand(['-section--special', 'abc'], 'section') // ['-section--special', 'abc']
 *
 */
export declare function expand(classList: string[], contextName: string): string[];
//# sourceMappingURL=class-list.d.ts.map