import type * as Style from '../index.js';
export declare function getLabelByName(name: string): string;
/**
 * Generate selectors that match without styling qualifier. Given class like `-card` it should be able match eleenets
 * with classes `-card`, `-card--test`, `test -card`, `test -card--test`, but not `test-card`
 *
 * @example
 *
 *   relaxClassName('-card') // ['[class^="-card"]', '[class*=" -card"]']
 *   relaxClassName('-block--inline') // ['[class^="-block--inline"]', '[class*=" -block--inline"]']
 *
 */
export declare function relaxClassName(className: string): string[];
/**
 * Get selectors used to match the definition. If definition disguises as another, it recurses to resolved target
 * definition.
 */
export declare function getDefinitionSelectors(definition: Style.ElementDefinition): string[];
/** Class is `-button`/`-button--style`, or `-block--badge`/`-block--badge--style` */
export declare function isElementClassName(className: string): boolean;
/**
 * Return element name implied by a class name.
 *
 * @example
 *
 *   getImplicitElementName('-section--test') // section
 *   getImplicitElementName('-button') // button
 *
 */
export declare function getImplicitElementName(className: string): "link" | "media" | "block" | "inline" | "blockquote" | "button" | "embed" | "form" | "section" | "time" | "var" | "paragraph" | "heading1" | "heading2" | "heading3" | "heading4" | "heading5" | "listItem" | "badge" | "card" | "component" | "spacer" | "container";
/**
 * Find element definition matching class name.
 *
 * @example
 *
 *   getDefinitionFromClassName('-section--cool') // Style.definition..section
 *   getDefinitionFromClassName('-block--test--style') // Style.definition..block
 *
 */
export declare function getDefinitionFromClassName(className: string): {
    readonly name: "paragraph";
    readonly id: "paragraph";
    readonly type: "text";
    readonly htmlTagName: "p";
    readonly className: "-paragraph";
    readonly disguise: string;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly placeholder: "Text element";
    readonly title: "Paragraphs";
    readonly label: "Paragraph";
    readonly description: "Regular text element used for all text and copy that isn't heading";
} | {
    readonly name: "heading1";
    readonly id: "heading1";
    readonly type: "text";
    readonly htmlTagName: "h1";
    readonly className: "-heading1";
    readonly disguise: string;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly title: "Heading 1";
    readonly label: "Heading 1";
    readonly description: "Main heading of the document, used for largest of headings";
} | {
    readonly name: "heading2";
    readonly id: "heading2";
    readonly type: "text";
    readonly htmlTagName: "h2";
    readonly className: "-heading2";
    readonly disguise: string;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly title: "Heading 2";
    /** Class is `-button`/`-button--style`, or `-block--badge`/`-block--badge--style` */
    readonly label: "Heading 2";
    readonly description: "Second level of headings";
} | {
    readonly name: "heading3";
    readonly id: "heading3";
    readonly type: "text";
    readonly htmlTagName: "h3";
    readonly className: "-heading3";
    readonly disguise: string;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly title: "Heading 3";
    readonly label: "Heading 3";
    readonly description: "Third level of headings often used for taglines";
} | {
    readonly name: "heading4";
    readonly id: "heading4";
    readonly type: "text";
    readonly htmlTagName: "h4";
    readonly className: "-heading4";
    readonly disguise: string;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly title: "Heading 4";
    readonly label: "Heading 4";
    readonly description: "Further level of headings for complex document";
} | {
    readonly name: "heading5";
    readonly id: "heading5";
    readonly type: "text";
    readonly htmlTagName: "h5";
    readonly className: "-heading5";
    readonly disguise: string;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly title: "Heading 5";
    readonly label: "Heading 5";
    readonly description: "The deepest heading level available";
} | {
    readonly name: "listItem";
    readonly id: "listItem";
    readonly type: "text";
    readonly htmlTagName: "li";
    readonly label: "List item";
    readonly disguise: "paragraph";
    readonly className: "-list-item";
    readonly isImplicit: true;
    readonly isCustom: false;
} | {
    readonly name: "link";
    readonly id: "link";
    readonly type: "inline";
    readonly label: "Link";
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly className: "-link";
    readonly disguise: string;
    readonly placeholder: "New link";
    readonly isImplicit: true;
    readonly canBeLink: true;
    readonly isCustom: false;
    readonly attributes: {
        readonly linkHref: "";
    };
    readonly title: "Links";
    readonly description: "Links inside text elements";
} | {
    readonly name: "button";
    readonly id: "button";
    readonly type: "inline";
    readonly label: "Button";
    readonly htmlTagName: "button";
    readonly className: "-button";
    readonly disguise: string;
    readonly placeholder: "Click me";
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly canBeLink: true;
    readonly title: "Buttons";
    readonly description: "Clickable element used for user interactions";
} | {
    readonly name: "badge";
    readonly id: "badge";
    readonly type: "inline";
    readonly label: "Badge";
    readonly className: "-inline--badge";
    readonly disguise: string;
    readonly placeholder: "Badge text";
    readonly isImplicit: false;
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly isCustom: true;
    readonly canBeLink: true;
    readonly title: "Badges";
    readonly description: "Element represeting status and other meta information";
} | {
    readonly name: "time";
    readonly id: "time";
    readonly type: "inline";
    readonly label: "Date";
    readonly htmlTagName: "time";
    readonly className: "-inline--time";
    readonly disguise: string;
    readonly placeholder: "2023-01-01";
    readonly isImplicit: true;
    readonly isCustom: true;
    readonly canBeLink: true;
    readonly description: "Present mapped date in human readable format";
} | {
    readonly name: "inline";
    readonly id: "inline";
    readonly type: "inline";
    readonly label: "Generic inline element";
    readonly className: "-inline";
    readonly disguise: string;
    readonly isImplicit: false;
    readonly placeholder: "Inline element";
    readonly canBeLink: true;
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly isCustom: false;
} | {
    readonly type: "inline";
    readonly id: "var";
    readonly label: "Variable";
    readonly name: "var";
    readonly className: string;
    readonly placeholder: "Double click to edit";
    readonly disguise: string;
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly description: "Inline placement to be substituted with mapped text";
    readonly attributes: {
        readonly 'data-path': "";
    };
} | {
    readonly name: "section";
    readonly id: "section";
    readonly type: "block";
    readonly label: "Section";
    readonly htmlTagName: "section";
    readonly className: "-section";
    readonly disguise: string;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly title: "Sections";
    readonly description: "Themeable essential boxes represeting top-level chunks of a component";
    readonly priority: -1;
    readonly attributes: {
        readonly class: "-grid--custom";
    };
} | {
    readonly name: "card";
    readonly id: "card";
    readonly type: "block";
    readonly label: "Card";
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly className: "-card";
    readonly disguise: string;
    readonly isImplicit: true;
    readonly canBeLink: true;
    readonly isCustom: false;
    readonly title: "Cards";
    readonly description: "Themable nestable boxes that provide visual framing for content";
    readonly priority: -1;
} | {
    readonly name: "blockquote";
    readonly id: "blockquote";
    readonly type: "block";
    readonly label: "Block quote";
    readonly htmlTagName: "blockquote";
    readonly className: "-blockquote";
    readonly disguise: string;
    readonly isImplicit: true;
    readonly canBeLink: true;
    readonly isCustom: false;
    readonly title: "Block quotes";
    readonly description: "Quotes, tweets and rich styled text references";
    readonly priority: -1;
} | {
    readonly name: "form";
    readonly id: "form";
    readonly type: "block";
    readonly label: "Form";
    readonly group: "external";
    readonly elementName: "component";
    readonly className: "-component--form";
    readonly disguise: "media";
    readonly canBeLink: false;
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly isImplicit: true;
    readonly isCustom: true;
    readonly description: "Form created in Sitecore Forms";
    readonly priority: 2;
} | {
    readonly name: "component";
    readonly id: "component";
    readonly type: "block";
    readonly label: "Component";
    readonly group: "external";
    readonly className: "-component";
    readonly disguise: "media";
    readonly canBeLink: false;
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly description: "Nested web, react or FEAAS component";
    readonly priority: 1;
} | {
    readonly name: "embed";
    readonly id: "embed";
    readonly type: "block";
    readonly group: "block";
    readonly label: "HTML Block";
    readonly className: "-embed";
    readonly disguise: "media";
    readonly canBeLink: true;
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly description: "Raw HTML content embedded into the component";
    readonly priority: 3;
} | {
    readonly name: "media";
    readonly id: "media";
    readonly type: "block";
    readonly label: "Image";
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly canBeLink: true;
    readonly className: "-block--media";
    readonly disguise: string;
    readonly title: "Image blocks";
    readonly description: "Image with optional accompanying text content";
    readonly priority: -1;
    readonly attributes: {
        readonly image: "";
    };
} | {
    readonly name: "block";
    readonly id: "block";
    readonly type: "block";
    readonly label: "Generic block element";
    readonly className: "-block";
    readonly disguise: string;
    readonly isImplicit: false;
    readonly canBeLink: true;
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly isCustom: false;
    readonly priority: 1;
} | {
    readonly name: "spacer";
    readonly id: "spacer";
    readonly type: "block";
    readonly label: "Spacer";
    readonly className: "-spacer";
    readonly disguise: string;
    readonly canBeLink: false;
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly description: "Flexible spacer that expands to take all available space";
    readonly priority: 2;
} | {
    readonly name: "container";
    readonly id: "container";
    readonly type: "block";
    readonly label: "Block";
    readonly className: "-container";
    readonly disguise: string;
    readonly canBeLink: true;
    readonly isImplicit: true;
    readonly isCustom: false;
    readonly htmlTagName: keyof HTMLElementTagNameMap;
    readonly priority: -1;
    readonly description: "Unstyled container for layout purposes";
};
export declare function matchClassName(className: string, candidate: string): boolean;
/**
 * Remove style qualifier from a qualifier, this will effectively revert element back to theme style.
 *
 * @example
 *
 *   getElementGenericClassName('-button') // '-button'
 *   getElementGenericClassName('-section--test') // '-section'
 *   getElementGenericClassName('-inline--chip--big') // '-inline--chip'
 *
 */
export declare function getElementGenericClassName(className: string): string;
/**
 * Returns element style type by class name (text/block/inline)
 *
 * @example
 *
 *   getElementType('-section--test') // 'block'
 *   getElementType('-heading1') // 'text'
 *   getElementType('-inline--badge--small') // 'inline'
 *
 */
export declare function getElementType(className: string): Style.Type.Element;
//# sourceMappingURL=conventions.d.ts.map