import type * as Style from './index.js';
export * from './methods/set.js';
/**
 * ## Style.Set -- manipulating sets of styles
 *
 * Style.Set is array of style rules. Methods in this module provide different ways to look the styles up. It's simple
 * things like finding styles by id, slug and the like, but also the domain-specific lookups like determining which
 * style is currently active for a context element.
 */
/**
 * Run fixers for all styles in a set.
 *
 * @param {Style.Rule[]} rules  - The array of style rules.
 * @returns {Style.Rule[]} The array of style rules after running fixers.
 * @example
 *
 *   const fixedRules = fix(originalRules)
 *   const refixedRules = fix(fixedRules) // refixedRules should be equal to fixedRules.
 *
 */
export declare function fix(rules: Style.Rule[]): Style.Rule[];
/**
 * List all choices of specific aspect type for given element.
 *
 * @param {Style.Rule[]}             rules  - The array of style rules.
 * @param {Style.Rule.Element}       style  - The element style.
 * @param {Style.Type.ElementAspect} type   - The element aspect type.
 * @returns {Style.Rule[]} The array of aspect choices for the given element.
 * @example
 *
 *   const decorationAspectChoices = getElementAspectChoices(rules, elementStyle, 'decoration')
 *   const typographyAspectChoices = getElementAspectChoices(rules, anotherElementStyle, 'typography')
 *
 */
export declare function getElementAspectChoices<S extends Style.Rule.Element, T extends Style.Type.ElementAspect<Style.typeOf<S>>>(rules: Style.Rule[], style: S, type: T): Style.Rule<T>[];
/**
 * For each possible aspect, list all choices.
 *
 * @param {Style.Rule[]}       rules    - The array of style rules.
 * @param {Style.Rule.Element} element  - The element style.
 * @returns {Object} The aspects by type.
 * @example
 *
 *   const headerAspectsByType = getElementAspectChoicesByType(rules, headerElementStyle)
 *   const paragraphAspectsByType = getElementAspectChoicesByType(rules, paragraphElementStyle)
 *
 */
export declare function getElementAspectChoicesByType<S extends Style.Rule.Element, T extends Style.Type.Element = Style.typeOf<S>>(rules: Style.Rule[], element: S): Style.AspectsByType<T>;
/**
 * Get element's assigned aspect.
 *
 * @param {Style.Rule[]}             rules    - The array of style rules.
 * @param {Style.ContextOrClassList} context  - The context or class list.
 * @param {Style.Type.Aspect}        type     - The aspect type.
 * @returns {Style.Rule} The assigned aspect for the given element.
 * @example
 *
 *   const assignedDecorationAspect = getContextAssignedAspect(rules, context, 'decoration')
 *   const assignedTypographyAspect = getContextAssignedAspect(rules, context, 'typography')
 *
 */
export declare function getContextAssignedAspect<T extends Style.Type.Aspect>(rules: Style.Rule[], context: Style.ContextOrClassList, type: T): Style.Rule<T>;
/**
 * Get context default aspect as defined by element style.
 *
 * @param {Style.Rule[]}             rules    - The array of style rules.
 * @param {Style.ContextOrClassList} context  - The context or class list.
 * @param {Style.Type.Aspect}        type     - The aspect type.
 * @returns {Style.Rule} The default aspect for the given context.
 * @example
 *
 *   const defaultDecorationAspect = getContextDefaultAspect(rules, context, 'decoration')
 *   const defaultTypographyAspect = getContextDefaultAspect(rules, context, 'typography')
 *
 */
export declare function getContextDefaultAspect<T extends Style.Type.Aspect>(rules: Style.Rule[], context: Style.ContextOrClassList, type: T): Style.Rule<T>;
/**
 * Get assigned context aspect in set of custom styles.
 *
 * @example
 *
 *   const customDecorationAspect = getContextCustomAspect(customRules, context, 'decoration')
 *   const customTypographyAspect = getContextCustomAspect(customRules, context, 'typography')
 *
 */
export declare function getContextCustomAspect<T extends Style.Type.Aspect>(customRules: Style.Rule[], context: Style.Context, type: T): Style.Rule<T>;
/**
 * List all possible element rules that can be applicable to given context.
 *
 * @param {Style.Rule[]}             rules    - The array of style rules.
 * @param {Style.ContextOrClassList} context  - The context or class list.
 * @returns {Style.Rule.Element[]} The array of possible element rules for the given context.
 * @example
 *
 *   const headerElementChoices = getContextElementChoices(rules, headerContext)
 *   const paragraphElementChoices = getContextElementChoices(rules, paragraphContext)
 *
 */
export declare function getContextElementChoices(rules: Style.Rule[], context: Style.ContextOrClassList): Style.Rule.Element[];
/**
 * List all possible element rules that can be applicable to given element definition.
 *
 * @param {Style.Rule[]}            rules       - The array of style rules.
 * @param {Style.ElementDefinition} definition  - Element definition.
 * @returns {Style.Rule.Element[]} The array of possible element rules for the given context.
 * @example
 *
 *   const headerElementChoices = getContextElementChoices(rules, headerDefinition)
 *   const paragraphElementChoices = getContextElementChoices(rules, paragraphDefinition)
 *
 */
export declare function getDefinitionElementChoices(rules: Style.Rule[], definition: Style.ElementDefinition): Style.Rule.Element[];
/**
 * Get all available aspect choices as defined by context's element rule.
 *
 * @param {Style.Rule[]}             rules    - The array of style rules.
 * @param {Style.ContextOrClassList} context  - The context or class list.
 * @param {Style.Type.ElementAspect} type     - The element aspect type.
 * @returns {Style.Rule[]} The array of available aspect choices for the given context's element rule.
 * @example
 *
 *   const decorationChoices = getContextAspectChoices(rules, context, 'decoration')
 *   const typographyChoices = getContextAspectChoices(rules, context, 'typography')
 *
 */
export declare function getContextAspectChoices<T extends Style.Type.ElementAspect>(rules: Style.Rule[], context: Style.ContextOrClassList, type: T): Style.Rule<T>[];
/** Get assigned element rule for given context. Returns null if it's generic element */
export declare function getContextElement(rules: Style.Rule[], context: Style.ContextOrClassList): Style.Rule.Element;
/** Return element definition for given context */
export declare function getContextDefinition(rules: Style.Rule[], context: Style.ContextOrClassList, followDisguise?: boolean): Style.ElementDefinition<"block" | "inline" | "text">;
/** List styles that are referenced explicitly in class list */
export declare function getContextAssigned<S extends Style.Rule>(rules: S[], context: Style.ContextOrClassList): Extract<S, Style.Rule.ElementAspect>[];
/**
 * Ultimate aspect getter that returns aspect for given context wether it's:
 *
 * - Custom rule assigned to element
 * - Assigned to element but not custom
 * - Default element aspect
 * - Aspect as defined by assigned combo
 * - Aspect as defined by inherited theme combo
 * - Aspect as defined by default theme combo
 */
export declare function getContextAspect<T extends Style.Type.Aspect>(rules: Style.Rule[], context: Style.Context, type: T, themeContext: Style.ContextOrClassList, customRules: Style.Rule[]): Style.Rule<T>;
/** Find aspect style for given context (custom, assigned or default) */
export declare function getContextElementAspect<T extends Style.Type.Aspect>(rules: Style.Rule[], context: Style.Context, type: T, customRules: Style.Rule[]): Style.Rule<T>;
export declare function getContextAspects<T extends Style.Type.Aspect>(rules: Style.Rule[], context: Style.Context, themeContext: Style.ContextOrClassList, customRules: Style.Rule[]): { [key in T]: Style.Rule<key>; };
/** Get combo for given context (assigned or inherited from theme) */
export declare function getContextCombo(rules: Style.Rule[], context: Style.ContextOrClassList, themeContext: Style.ContextOrClassList): {
    id: string;
    refId: string;
    fillId: string;
    decorationId: string;
    spacingId: string;
    typographyId: string;
    paletteId: string;
    isDefault: boolean;
};
/** Get theme for given context */
export declare function getContextTheme(rules: Style.Rule[], context: Style.ContextOrClassList): Style.RuleWithProps<"theme", {
    blocks: {
        id: string;
        refId: string;
        fillId: string;
        decorationId: string;
        spacingId: string;
        typographyId: string;
        paletteId: string;
        isDefault: boolean;
    }[];
    inlines: {
        id: string;
        refId: string;
        fillId: string;
        decorationId: string;
        spacingId: string;
        typographyId: string;
        paletteId: string;
        isDefault: boolean;
    }[];
    texts: {
        id: string;
        refId: string;
        fillId: string;
        decorationId: string;
        spacingId: string;
        typographyId: string;
        paletteId: string;
        isDefault: boolean;
    }[];
}>;
/** Get combo explicitly assigned to given context */
export declare function getContextAssignedCombo(rules: Style.Rule[], context: Style.ContextOrClassList): {
    id: string;
    refId: string;
    fillId: string;
    decorationId: string;
    spacingId: string;
    typographyId: string;
    paletteId: string;
    isDefault: boolean;
};
/** Get theme combo matching given context */
export declare function getContextThemeCombo(rules: Style.Rule[], context: Style.ContextOrClassList, themeContext: Style.ContextOrClassList): {
    id: string;
    refId: string;
    fillId: string;
    decorationId: string;
    spacingId: string;
    typographyId: string;
    paletteId: string;
    isDefault: boolean;
};
/** Get theme combo that is chosen as default that will match given context element */
export declare function getContextThemeDefaultCombo(rules: Style.Rule[], context: Style.ContextOrClassList, themeContext: Style.ContextOrClassList): {
    id: string;
    refId: string;
    fillId: string;
    decorationId: string;
    spacingId: string;
    typographyId: string;
    paletteId: string;
    isDefault: boolean;
};
/** Get aspect specified by combo for given context */
export declare function getContextComboAspect<T extends Style.Type.ElementAspect>(rules: Style.Rule[], context: Style.ContextOrClassList, type: T, themeContext: Style.ContextOrClassList): Style.Rule<T>;
/** Given a set of element choices, return all matching combos in a theme */
export declare function getElementsThemeCombos(elementChoices: Style.Rule[], theme: Style.Rule<'theme'>): {
    id: string;
    refId: string;
    fillId: string;
    decorationId: string;
    spacingId: string;
    typographyId: string;
    paletteId: string;
    isDefault: boolean;
}[];
/** Analyze the level of customization for given element rule */
export declare function getElementCustomizationType<S extends Style.Rule.Element, T extends Style.typeOf<S>>(effectiveRules: Style.Rule[], element: S, assignedRules: Style.Rule.ElementAspect<T>[]): "oneOff" | "default" | "allowed" | "nonAllowed";
export declare function getDefinitionCustomizationType<S extends Style.Rule.Element, T extends Style.typeOf<S>>(rules: Style.Rule[], definition: Style.ElementDefinition, assignedRules: Style.Rule.ElementAspect<T>[]): "oneOff" | "default" | "allowed" | "nonAllowed";
//# sourceMappingURL=set.d.ts.map